@@ -7,7 +7,7 @@
 
 /***********************************************************************
  *
- * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2006
+ * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2006,2007,2008
  * Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
@@ -224,12 +224,26 @@ typedef struct gst_block_closure
 }
  *gst_block_closure;
 
+typedef struct compiler_state
+{
+  int line_offset;
+  int prev_line;
+  scope cur_scope;
+  mst_Boolean inside_block;
+  mst_Boolean undeclared_temporaries;
+  OOP *literal_vec;
+  OOP *literal_vec_curr;
+  OOP *literal_vec_max;
+  jmp_buf bad_method;
+} compiler_state;
 
-/* These hold the compiler's notions of the current class for compilations,
-   and the current category that compiled methods are to be placed into */
-extern OOP _gst_this_class ATTRIBUTE_HIDDEN;
-extern OOP _gst_this_category ATTRIBUTE_HIDDEN;
-extern OOP _gst_latest_compiled_method ATTRIBUTE_HIDDEN;
+
+/* This holds the parse tree for the method being compiled.  */
+extern tree_node _gst_curr_method ATTRIBUTE_HIDDEN;
+
+/* This holds other bits of compiler state that needs to be saved
+   when the compiler is reentered.  */
+extern compiler_state *_gst_compiler_state ATTRIBUTE_HIDDEN;
 
 /* This is the value most recently returned by
    _gst_execute_statements.  It is used to communicate the returned
@@ -253,11 +267,6 @@ extern int _gst_declare_tracing
 extern mst_Boolean _gst_skip_compilation 
   ATTRIBUTE_HIDDEN;
 
-/* This holds whether the compiler should make the compiled methods
-   untrusted.  */
-extern mst_Boolean _gst_untrusted_methods 
-  ATTRIBUTE_HIDDEN;
-
 /* Given a tree_node of type TREE_KEYWORD_LIST, pick out and concatenate
    the keywords, turn them into a symbol OOP and return that symbol.  */
 extern OOP _gst_compute_keyword_selector (tree_node selectorExpr)
@@ -270,23 +279,14 @@ extern OOP _gst_compute_keyword_selector (tree_node selectorExpr)
 extern OOP _gst_make_constant_oop (tree_node constExpr);
 
 /* Called to compile and execute an "immediate expression"; i.e. a Smalltalk
-   statement that is not part of a method definition and where temporaries are
-   declared automatically.  The parse trees are in TEMPS and STATEMENTS.
-   Return the object that was returned by the expression.  */
-extern OOP _gst_execute_statements (tree_node temps,
-				    tree_node statement,
-				    enum undeclared_strategy undeclared,
+   statement that is not part of a method definition.  The parse trees are in
+   METHOD.  Return the object that was returned by the expression.  */
+extern OOP _gst_execute_statements (OOP receiverOOP,
+				    tree_node method,
+				    mst_Boolean undeclaredTemps,
 				    mst_Boolean quiet)
   ATTRIBUTE_HIDDEN;
 
-/* This function will print a message describing the method category
-   and class being compiled.  The message can have the form "STRING
-   _GST_THIS_CATEGORY for _GST_THIS_CLASS" or "STRING for
-   _GST_THIS_CLASS", depending on the value of CATEGORY.  */
-extern void _gst_display_compilation_trace (const char *string,
-				            mst_Boolean category) 
-  ATTRIBUTE_HIDDEN;
-
 /* This routine does a very interesting thing.  It installs the inital
    method, which is the primitive for "methodsFor:".  It does this by
    creating a string that contains the method definition and then
@@ -303,17 +303,9 @@ extern void _gst_display_compilation_trace (const char *string,
 extern void _gst_install_initial_methods (void) 
   ATTRIBUTE_HIDDEN;
 
-/* Sets the compiler's notion of the class to compile methods into.  */
-extern void _gst_set_compilation_class (OOP class_oop) 
-  ATTRIBUTE_HIDDEN;
-
-/* Sets the compiler's notion of the category to compile methods into.  */
-extern void _gst_set_compilation_category (OOP categoryOOP) 
-  ATTRIBUTE_HIDDEN;
-
 /* Clears the compiler's notion of the class and category to compile
    methods into.  */
-extern void _gst_reset_compilation_category () 
+extern void reset_compilation_category () 
   ATTRIBUTE_HIDDEN;
 
 /* This function will send a message to ObjectMemory (a system
@@ -321,10 +313,6 @@ extern void _gst_reset_compilation_category ()
 extern void _gst_invoke_hook (enum gst_vm_hook hook) 
   ATTRIBUTE_HIDDEN;
 
-/* Prepares the compiler for execution, initializing some variables.  */
-extern void _gst_init_compiler (void) 
-  ATTRIBUTE_HIDDEN;
-
 /* Compile the code for a complete method definition.  This basically
    walks the METHOD parse tree, but in addition it special cases for
    methods that don't return a value explicitly by returning "self".
@@ -333,7 +321,8 @@ extern void _gst_init_compiler (void)
    derived from the method expression.  */
 extern OOP _gst_compile_method (tree_node method,
 				mst_Boolean returnLast,
-				mst_Boolean install) 
+				mst_Boolean install,
+                                mst_Boolean undeclaredTemps) 
   ATTRIBUTE_HIDDEN;
 
 /* Constructs and returns a new CompiledMethod instance.  It computes
@@ -341,10 +330,10 @@ extern OOP _gst_compile_method (tree_node method,
    the method's byte codes (setting up the flags to optimize returns).
    LITERALS is a Smalltalk Array containing the literals, or nil if we
    retrieve it from the array internal to comp.c */
-extern OOP _gst_make_new_method (int primitiveIndex,
-				 int numArgs,
+extern OOP _gst_make_new_method (int numArgs,
 				 int numTemps,
 				 int maximumStackDepth,
+				 OOP attributes,
 				 OOP literals,
 				 bc_vector bytecodes,
 				 OOP class,
@@ -380,15 +369,4 @@ extern OOP _gst_block_new (int numArgs,
 extern int _gst_add_forced_object (OOP oop) 
   ATTRIBUTE_HIDDEN;
 
-/* Transform the ATTRIBUTE_KEYWORDS node (a TREE_ATTRIBUTE_LIST)
-   into a Message object, and return it.  */
-extern OOP _gst_make_attribute (tree_node attribute_keywords) 
-  ATTRIBUTE_HIDDEN;
-
-/* Process the attributes in ARRAYOOP, return the primitive number
-   (so far, this is the only attribute we honor), or -1 for a bad
-   primitive number.  */
-extern int _gst_process_attributes_array (OOP arrayOOP) 
-  ATTRIBUTE_HIDDEN;
-
 #endif /* GST_COMP_H */
