@@ -7,7 +7,8 @@
 
 /***********************************************************************
  *
- * Copyright 2000, 2001, 2002, 2003, 2005, 2006, 2008 Free Software Foundation, Inc.
+ * Copyright 2000, 2001, 2002, 2003, 2005, 2006, 2008, 2009
+ * Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
  * This file is part of GNU Smalltalk.
@@ -124,7 +125,9 @@ typedef struct gst_identity_dictionary
 {
   OBJ_HEADER;
   OOP tally;			/* really, an int */
-  OOP keys[1];			/* variable sized array of OOPS (keys/values) */
+
+  /* Other, indexable fields that contain keys and values for this
+     dictionary.  */
 }
  *gst_identity_dictionary;
 
@@ -260,6 +263,15 @@ typedef struct gst_directed_message
 }
  *gst_directed_message;
 
+typedef struct gst_message_lookup
+{
+  OBJ_HEADER;
+  OOP selector;
+  OOP args;
+  OOP startingClass;
+}
+ *gst_message_lookup;
+
 typedef struct gst_string
 {
   OBJ_HEADER;
@@ -291,7 +303,6 @@ typedef struct gst_class
   OOP environment;
   OOP classVariables;		/* dictionary of name->value pairs */
   OOP sharedPools;
-  OOP securityPolicy;
   OOP pragmaHandlers;
 }
  *gst_class;
@@ -338,7 +349,6 @@ extern OOP _gst_abstract_namespace_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_array_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_arrayed_collection_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_association_class ATTRIBUTE_HIDDEN;
-extern OOP _gst_bag_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_behavior_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_binding_dictionary_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_block_closure_class ATTRIBUTE_HIDDEN;
@@ -364,9 +374,9 @@ extern OOP _gst_context_part_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_continuation_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_date_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_deferred_variable_binding_class ATTRIBUTE_HIDDEN;
-extern OOP _gst_delay_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_dictionary_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_directed_message_class ATTRIBUTE_HIDDEN;
+extern OOP _gst_message_lookup_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_false_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_file_descriptor_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_file_segment_class ATTRIBUTE_HIDDEN;
@@ -392,7 +402,6 @@ extern OOP _gst_linked_list_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_lookup_key_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_lookup_table_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_magnitude_class ATTRIBUTE_HIDDEN;
-extern OOP _gst_mapped_collection_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_memory_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_message_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_metaclass_class ATTRIBUTE_HIDDEN;
@@ -415,7 +424,6 @@ extern OOP _gst_security_policy_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_semaphore_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_sequenceable_collection_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_set_class ATTRIBUTE_HIDDEN;
-extern OOP _gst_shared_queue_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_small_integer_class ATTRIBUTE_HIDDEN;
 extern OOP _gst_smalltalk_dictionary ATTRIBUTE_HIDDEN;
 extern OOP _gst_sorted_collection_class ATTRIBUTE_HIDDEN;
@@ -580,20 +588,6 @@ extern OOP _gst_identity_dictionary_at_put (OOP identityDictionaryOOP,
 					    OOP valueOOP) 
   ATTRIBUTE_HIDDEN;
 
-/* Called when a dictionary becomes full, this routine replaces the
-   dictionary instance that DICTIONARYOOP is pointing to with a new,
-   larger dictionary, and returns this new dictionary (the object
-   pointer, not the OOP).  */
-extern gst_object _gst_grow_dictionary (OOP dictionaryOOP) 
-  ATTRIBUTE_HIDDEN;
-
-/* Called when an IdentityDictionary becomes full, this routine
-   replaces the IdentityDictionary instance that IDENTITYDICTIONARYOOP
-   is pointing to with a new, larger dictionary, and returns this new
-   dictionary (the object pointer, not the OOP).  */
-extern gst_identity_dictionary _gst_grow_identity_dictionary (OOP identityDictionaryOOP) 
-  ATTRIBUTE_HIDDEN;
-
 /* Allocates and returns a new C (NULL-terminated) string that has the same
    contents as STRINGOOP.  Even if there are embedded NULs, the
    allocated area has always a size of "stringOOP size + 1" bytes.  */
@@ -664,4 +658,8 @@ extern mst_Boolean _gst_init_dictionary_on_image_load (mst_Boolean prim_table_ma
 extern int _gst_resolve_primitive_name (char *name) 
   ATTRIBUTE_HIDDEN;
 
+/* Entry point for the profiler.  */
+extern void _gst_record_profile (OOP oldMethod, OOP newMethod, int ipOffset)
+  ATTRIBUTE_HIDDEN;
+
 #endif /* GST_DICT_H */
