@@ -162,6 +162,7 @@ static const char standard_files[] = {
   "SysDict.st\0"
   "Object.st\0"
   "Message.st\0"
+  "MessageLookup.st\0"
   "DirMessage.st\0"
   "Boolean.st\0"
   "False.st\0"
@@ -238,7 +239,6 @@ static const char standard_files[] = {
   "FileSegment.st\0"
   "FileDescr.st\0"
   "SymLink.st\0"
-  "Security.st\0"
   "WeakObjects.st\0"
   "ObjMemory.st\0"
 
@@ -286,7 +286,6 @@ static const char standard_files[] = {
   "Getopt.st\0"
   "Generator.st\0"
   "StreamOps.st\0"
-  "ObjDumper.st\0"
   "Regex.st\0"
   "PkgLoader.st\0"
   "Autoload.st\0"
@@ -456,6 +455,7 @@ _gst_initialize (const char *kernel_dir,
 
   _gst_init_sysdep ();
   _gst_init_signals ();
+  _gst_init_event_loop();
   _gst_init_cfuncs ();
   _gst_init_sockets ();
   _gst_init_primitives ();
@@ -496,7 +496,6 @@ _gst_initialize (const char *kernel_dir,
   if (loadBinary && _gst_load_from_file (_gst_binary_image_name))
     {
       _gst_init_interpreter ();
-      _gst_init_compiler ();
       _gst_init_vmproxy ();
     }
   else if (abortOnFailure)
@@ -514,7 +513,6 @@ _gst_initialize (const char *kernel_dir,
       _gst_init_mem_default ();
       _gst_init_dictionary ();
       _gst_init_interpreter ();
-      _gst_init_compiler ();
       _gst_init_vmproxy ();
 
       _gst_install_initial_methods ();
