@@ -94,6 +94,13 @@ static void recover_error (gst_parser *p)
 static int filprintf (Filament *fil,
 		      const char *format, ...);
 
+/* Transform the ATTRIBUTE_KEYWORDS node (a TREE_ATTRIBUTE_LIST)
+   into a Message object, and return it.  Compile the arguments
+   in the context of CLASSOOP if not NULL.  */
+static OOP make_attribute (gst_parser *p,
+                           OOP classOOP,
+			   tree_node attribute_keywords);
+
 /* Grammar productions.  */
 
 static void parse_chunks (gst_parser *p);
@@ -105,15 +112,18 @@ static mst_Boolean parse_scoped_definition (gst_parser *p,
 
 static void parse_eval_definition (gst_parser *p);
 
-static mst_Boolean parse_and_send_attribute (gst_parser *p,
-					     OOP receiverOOP);
+static void parse_and_send_attribute (gst_parser *p,
+                                      OOP receiverOOP,
+                                      mst_Boolean forClass);
 static mst_Boolean parse_namespace_definition (gst_parser *p,
 					       tree_node first_stmt);
 static mst_Boolean parse_class_definition (gst_parser *p,
 					   OOP classOOP,	
 					   mst_Boolean extend);
-static OOP parse_namespace (tree_node name);
-static OOP parse_class (tree_node list);
+static OOP parse_namespace (gst_parser *p,
+			    tree_node name);
+static OOP parse_class (gst_parser *p,
+			tree_node list);
 static void parse_scoped_method (gst_parser *p,
 				 OOP classOOP);
 static void parse_instance_variables (gst_parser *p,
@@ -198,7 +208,6 @@ lex_lookahead (gst_parser *p, int n)
   while (p->la_size < n)
     {
       int i = (p->la_first + p->la_size) % 4;
-      assert (p->la_size == 0 || token (p, p->la_size - 1) != EOF);
       p->la[i].token = _gst_yylex (&p->la[i].val, &p->la[i].loc);
       p->la_size++;
     }
@@ -306,40 +315,142 @@ _gst_print_tokens (gst_parser *p)
   printf ("\n");
 }
 
-/* Top of the descent.  */
-
-void
-_gst_parse_method ()
+OOP
+_gst_get_current_namespace (void)
 {
-  gst_parser p, *prev_parser = _gst_current_parser;
-  _gst_current_parser = &p;
-  p.state = PARSE_METHOD;
-  lex_init (&p);
-  if (setjmp (p.recover) == 0)
-    parse_method (&p, ']');
-  else
-    _gst_had_error = false;
+  if (_gst_current_parser)
+    return _gst_current_parser->current_namespace;
 
-  _gst_current_parser = prev_parser;
+  return _gst_current_namespace;
 }
 
 void
-_gst_parse_chunks ()
+_gst_set_compilation_class (OOP class_oop)
+{
+  assert (IS_NIL (_gst_current_parser->currentClass));
+  _gst_current_parser->currentClass = class_oop;
+  _gst_register_oop (class_oop);
+  _gst_compute_linearized_pools (_gst_current_parser, false);
+}
+
+void
+_gst_set_compilation_category (OOP categoryOOP)
+{
+  assert (IS_NIL (_gst_current_parser->currentCategory));
+  _gst_current_parser->currentCategory = categoryOOP;
+  _gst_register_oop (categoryOOP);
+}
+
+static OOP
+set_compilation_namespace (OOP namespaceOOP)
+{
+  OOP oldNamespaceOOP = _gst_current_parser->current_namespace;
+
+  _gst_current_parser->current_namespace = namespaceOOP;
+  if (!IS_NIL (namespaceOOP))
+    {
+      if (oldNamespaceOOP != namespaceOOP)
+        _gst_register_oop (namespaceOOP);
+      if (namespaceOOP != _gst_current_namespace)
+        _gst_msg_sendf (NULL, "%v %o current: %o",
+                        _gst_namespace_class, namespaceOOP);
+    }
+
+  if (!IS_NIL (oldNamespaceOOP))
+    {
+      if (oldNamespaceOOP != namespaceOOP)
+        _gst_unregister_oop (_gst_current_parser->current_namespace);
+    }
+
+  return oldNamespaceOOP;
+}
+
+void
+_gst_reset_compilation_category (void)
+{
+  if (!IS_NIL (_gst_current_parser->currentClass))
+    {
+      _gst_unregister_oop (_gst_current_parser->currentClass);
+      _gst_current_parser->currentClass = _gst_nil_oop;
+      _gst_free_linearized_pools ();
+    }
+  if (!IS_NIL (_gst_current_parser->currentCategory))
+    {
+      _gst_unregister_oop (_gst_current_parser->currentCategory);
+      _gst_current_parser->currentCategory = _gst_nil_oop;
+    }
+}
+
+static void
+parser_init (gst_parser *p)
+{
+  memset (p, 0, sizeof(*p));
+  p->currentClass = p->currentCategory = _gst_nil_oop;
+  /* This should ultimately become _gst_get_current_namespace ().  */
+  p->current_namespace = _gst_current_namespace;
+}
+
+/* Top of the descent.  */
+
+OOP
+_gst_parse_method (OOP currentClass, OOP currentCategory)
 {
   gst_parser p, *prev_parser = _gst_current_parser;
+  OOP methodOOP;
+  inc_ptr incPtr;
 
   _gst_current_parser = &p;
+  incPtr = INC_SAVE_POINTER ();
+  parser_init (&p);
+  p.state = PARSE_METHOD;
+  p.current_namespace = _gst_nil_oop;
+  _gst_set_compilation_class (currentClass);
+  _gst_set_compilation_category (currentCategory);
+
+  lex_init (&p);
+  if (setjmp (p.recover) == 0)
+    {
+      parse_method (&p, ']');
+      methodOOP = p.lastMethodOOP;
+    }
+  else
+    {
+      _gst_had_error = false;
+      methodOOP = _gst_nil_oop;
+    }
+
+  _gst_reset_compilation_category ();
+  set_compilation_namespace (_gst_nil_oop);
+  _gst_current_parser = prev_parser;
+  INC_RESTORE_POINTER (incPtr);
+  return methodOOP;
+}
+
+void
+_gst_parse_chunks (OOP currentNamespace)
+{
+  gst_parser p, *prev_parser = _gst_current_parser;
+  inc_ptr incPtr;
+
+  _gst_current_parser = &p;
+  incPtr = INC_SAVE_POINTER ();
+  parser_init (&p);
+  if (currentNamespace)
+    p.current_namespace = currentNamespace;
+  p.state = PARSE_DOIT;
 
   lex_init (&p);
   if (token (&p, 0) == SHEBANG)
     lex (&p);
 
-  p.state = PARSE_DOIT;
   setjmp (p.recover);
   _gst_had_error = false;
   while (token (&p, 0) != EOF)
     parse_chunks (&p);
 
+  _gst_reset_compilation_category ();
+  set_compilation_namespace (_gst_nil_oop);
+  INC_RESTORE_POINTER (incPtr);
   _gst_current_parser = prev_parser;
 }
 
@@ -352,7 +463,6 @@ parse_chunks (gst_parser *p)
     {
       OOP oldTemporaries = _gst_push_temporaries_dictionary ();
       jmp_buf old_recover;
-      memcpy (old_recover, p->recover, sizeof (p->recover));
       setjmp (p->recover);
       while (token (p, 0) != EOF && token (p, 0) != '!')
         {
@@ -440,6 +550,57 @@ recover_error (gst_parser *p)
   longjmp (p->recover, 1);
 }
 
+static OOP
+execute_doit (gst_parser *p, tree_node temps, tree_node stmts,
+              OOP receiverOOP,
+	      mst_Boolean undeclared, mst_Boolean quiet)
+{
+  tree_node method;
+  mst_Boolean in_class;
+  OOP resultOOP;
+  inc_ptr incPtr;
+
+  in_class = !IS_NIL (p->currentClass);
+  if (!in_class)
+    {
+      assert (p == _gst_current_parser);
+      if (receiverOOP)
+        _gst_set_compilation_class (OOP_CLASS (receiverOOP));
+      else
+        {
+          /* Let doits access the variables and classes in the current namespace.  */
+          set_compilation_namespace (p->current_namespace);
+	  receiverOOP = _gst_nil_oop;
+          p->currentClass = _gst_undefined_object_class;
+          _gst_register_oop (p->currentClass);
+          _gst_compute_linearized_pools (p, true);
+        }
+    }
+
+  if (stmts->nodeType != TREE_STATEMENT_LIST)
+    stmts = _gst_make_statement_list (&stmts->location, stmts);
+
+  method =
+    _gst_make_method (&stmts->location, loc(p, 0),
+                      NULL, temps, NULL, stmts,
+                      _gst_get_current_namespace (),
+                      _gst_current_parser->currentClass,
+                      _gst_nil_oop,
+                      false);
+
+  resultOOP = _gst_execute_statements (receiverOOP, method, undeclared, quiet);
+  incPtr = INC_SAVE_POINTER ();
+  INC_ADD_OOP (resultOOP);
+
+  if (!in_class && p->state != PARSE_METHOD_LIST)
+    _gst_reset_compilation_category ();
+
+  /* Store Smalltalk's notion of current namespace back to the parser.  */
+  set_compilation_namespace (_gst_current_namespace);
+  INC_RESTORE_POINTER (incPtr);
+  return resultOOP;
+}
+
 /* doit: temporaries statements '!' [ method_list '!' ]
    | empty */
 
@@ -469,8 +630,13 @@ parse_doit (gst_parser *p, mst_Boolean fail_at_eof)
     }
   else if (statement)
     {
-      _gst_execute_statements (NULL, statement, UNDECLARED_TEMPORARIES, false);
-      _gst_free_tree ();
+      execute_doit (p, NULL, statement, NULL, true, false);
+
+      /* Because a '.' could be inserted automagically, the next token
+         value might be already on the obstack.  Do not free in that
+         case!  */
+      if (p->la_size == 0)
+        _gst_free_tree ();
     }
 
   _gst_had_error = false;
@@ -512,10 +678,11 @@ parse_scoped_definition (gst_parser *p, tree_node first_stmt)
 	return parse_namespace_definition (p, first_stmt);
       
       if (strcmp (expression->v_list.name, "subclass:") == 0
-	  && (classOOP = parse_class (receiver)) != NULL)
+	  && (classOOP = parse_class (p, receiver)) != NULL)
 	{
 	  const char * name = expression->v_list.value->v_list.name;
-	  _gst_msg_sendf (&classOOP, "%o %o subclass: %S", classOOP, name);
+	  _gst_msg_sendf (&classOOP, "%o %o subclass: %S environment: %o",
+                          classOOP, name, p->current_namespace);
 	      
 	  if (IS_NIL (classOOP))
 	    _gst_had_error = true;
@@ -527,38 +694,29 @@ parse_scoped_definition (gst_parser *p, tree_node first_stmt)
   else if (first_stmt->nodeType == TREE_UNARY_EXPR
 	   && first_stmt->v_expr.selector == _gst_intern_string ("extend"))
     {
-      OOP namespace_old = _gst_current_namespace;
       OOP classOrMetaclassOOP = NULL;
       mst_Boolean ret_value;
 
-      _gst_register_oop (namespace_old);
       if (receiver->nodeType == TREE_VARIABLE_NODE)
  	{
- 	  classOOP = parse_class (receiver);
+ 	  classOOP = parse_class (p, receiver);
 	  classOrMetaclassOOP = classOOP;
  	}
       else if (receiver->nodeType == TREE_UNARY_EXPR
 	       && receiver->v_expr.selector == _gst_intern_string ("class"))
  	{
- 	  classOOP = parse_class (receiver->v_expr.receiver);
+ 	  classOOP = parse_class (p, receiver->v_expr.receiver);
 	  classOrMetaclassOOP = classOOP ? OOP_CLASS (classOOP) : NULL;
  	}	   
       if (classOrMetaclassOOP != NULL) 
  	{
- 	  OOP namespace_new = ((gst_class) OOP_TO_OBJ (classOOP))->environment; 
-	  
-	  /* When creating the image, current namespace is not available. */
-	  if (namespace_new != namespace_old)
-	    _gst_msg_sendf (NULL, "%v %o current: %o",
-			    _gst_namespace_class, namespace_new);
-	  
+          inc_ptr incPtr = INC_SAVE_POINTER ();
+ 	  OOP namespace_new = ((gst_class) OOP_TO_OBJ (classOOP))->environment;
+          OOP namespace_old = set_compilation_namespace (namespace_new);
+          INC_ADD_OOP (namespace_old);
  	  ret_value = parse_class_definition (p, classOrMetaclassOOP, true);
-
-	  if (namespace_new != namespace_old)
-	    _gst_msg_sendf (NULL, "%v %o current: %o",
-			    _gst_namespace_class, namespace_old);
-	  
- 	  _gst_unregister_oop (namespace_old);
+          set_compilation_namespace (namespace_old);
+ 	  INC_RESTORE_POINTER (incPtr);
  	  return ret_value;
  	}
     }
@@ -575,15 +733,18 @@ parse_eval_definition (gst_parser *p)
   OOP oldDictionary = _gst_push_temporaries_dictionary ();
   jmp_buf old_recover;
 
+  assert (IS_NIL (_gst_current_parser->currentClass));
   memcpy (old_recover, p->recover, sizeof (p->recover));
   if (setjmp (p->recover) == 0)
     {
       tmps = parse_temporaries (p, false);
       stmts = parse_statements (p, NULL, true);
+      lex_must_be (p, ']');
     }
 
   if (stmts && !_gst_had_error)
     {
+      OOP resultOOP;
       if (_gst_regression_testing)
         {
           printf ("\nExecution begins...\n");
@@ -591,13 +752,13 @@ parse_eval_definition (gst_parser *p)
 	  fflush (stderr);
         }
 
-      _gst_execute_statements (tmps, stmts, UNDECLARED_TEMPORARIES,
-			       _gst_regression_testing);
+      resultOOP = execute_doit (p, tmps, stmts, NULL, true,
+                                _gst_regression_testing);
 
       if (_gst_regression_testing)
         {
           if (!_gst_had_error)
-            printf ("returned value is %O\n", _gst_last_returned_value);
+            printf ("returned value is %O\n", resultOOP);
 	  fflush (stdout);
 	  fflush (stderr);
         }
@@ -605,6 +766,7 @@ parse_eval_definition (gst_parser *p)
       _gst_had_error = false;
     }
 
+  assert (p->la_size <= 1);
   _gst_free_tree ();
   _gst_pop_temporaries_dictionary (oldDictionary);
   memcpy (p->recover, old_recover, sizeof (p->recover));
@@ -612,41 +774,73 @@ parse_eval_definition (gst_parser *p)
     longjmp (p->recover, 1);
 }
 
-static mst_Boolean
-parse_and_send_attribute (gst_parser *p, OOP receiverOOP)
+static OOP
+make_attribute (gst_parser *p, OOP classOOP, tree_node attribute_keywords)
 {
-  OOP selectorOOP, *args;
-  tree_node keyword, value, stmt;
-  int i, nb = 0;
+  tree_node keyword;
+  OOP selectorOOP, argsArrayOOP, messageOOP;
+  gst_object argsArray;
+  int i, numArgs;
+  inc_ptr incPtr;
 
-#if 0
-  printf ("parse attribute\n");
-#endif
-  lex_skip_mandatory (p, '<');
-  keyword = parse_keyword_expression (p, NULL, EXPR_KEYWORD);
+  incPtr = INC_SAVE_POINTER ();
+  
+  if (_gst_had_error)
+    return _gst_nil_oop;
 
-  selectorOOP = _gst_compute_keyword_selector (keyword->v_expr.expression);
-  nb = _gst_selector_num_args (selectorOOP);
-  args = alloca (sizeof (*args) * nb);
-  i = 0;
-  for (stmt = keyword->v_expr.expression; stmt; stmt = stmt->v_list.next)
+  selectorOOP = _gst_compute_keyword_selector (attribute_keywords);
+  numArgs = _gst_list_length (attribute_keywords);
+  argsArray = instantiate_with (_gst_array_class, numArgs, &argsArrayOOP);
+  INC_ADD_OOP (argsArrayOOP);
+
+  for (i = 0, keyword = attribute_keywords; keyword != NULL;
+       i++, keyword = keyword->v_list.next)
     {
-      value = stmt->v_list.value;
-      value = _gst_make_statement_list (&value->location, value);
-      args[i] = _gst_execute_statements (NULL, value, UNDECLARED_NONE, true);
-      if (!args[i])
+      tree_node value = keyword->v_list.value;
+      OOP result = execute_doit (p, NULL, value, classOOP, false, true);
+      if (!result)
         {
           _gst_had_error = true;
-          break;
+          INC_RESTORE_POINTER (incPtr);
+          return _gst_nil_oop;
         }
-      i = i + 1;
+
+      argsArray = OOP_TO_OBJ (argsArrayOOP);
+      argsArray->data[i] = result;
     }
 
+  messageOOP = _gst_message_new_args (selectorOOP, argsArrayOOP);
+  INC_RESTORE_POINTER (incPtr);
+
+  MAKE_OOP_READONLY (argsArrayOOP, true);
+  MAKE_OOP_READONLY (messageOOP, true);
+  return (messageOOP);
+}
+
+
+static void
+parse_and_send_attribute (gst_parser *p, OOP receiverOOP, mst_Boolean forClass)
+{
+  tree_node keywords;
+
+  lex_skip_mandatory (p, '<');
+  keywords = parse_keyword_list (p, EXPR_BINOP);
   if (!_gst_had_error)
-    _gst_nvmsg_send (receiverOOP, selectorOOP, args, i);
+    {
+      OOP messageOOP;
+      messageOOP = make_attribute (p, forClass ? receiverOOP : _gst_nil_oop, keywords);
+      if (!IS_NIL (messageOOP))
+        {
+          OOP selectorOOP = MESSAGE_SELECTOR (messageOOP);
+          OOP argsOOP = MESSAGE_ARGS (messageOOP);
+	  int numArgs = NUM_OOPS (OOP_TO_OBJ (argsOOP));
+          OOP *args = alloca(numArgs * sizeof (OOP));
+          memcpy (args, OOP_TO_OBJ (argsOOP)->data, numArgs * sizeof (OOP));
+          _gst_nvmsg_send (receiverOOP, selectorOOP, args, numArgs);
+        }
+    }
 
   lex_skip_mandatory (p, '>');
-  return !_gst_had_error;
 }
 
 
@@ -654,28 +848,23 @@ static mst_Boolean
 parse_namespace_definition (gst_parser *p, tree_node first_stmt)
 {      
   tree_node expr = first_stmt->v_expr.expression;
-  OOP new_namespace = parse_namespace (expr->v_list.value);
+  OOP new_namespace = parse_namespace (p, expr->v_list.value);
   
   if (new_namespace)
     {
-      OOP old_namespace = _gst_current_namespace;
-      _gst_register_oop (old_namespace);		
-  
-      _gst_msg_sendf (NULL, "%v %o current: %o", 
-		      _gst_namespace_class, new_namespace);
-      
+      inc_ptr incPtr = INC_SAVE_POINTER ();
+      OOP old_namespace = set_compilation_namespace (new_namespace);
+      INC_ADD_OOP (old_namespace);
       while (token (p, 0) != ']' && token (p, 0) != EOF && token (p, 0) != '!')
 	{
 	  if (token (p, 0) == '<')
-	    parse_and_send_attribute (p, new_namespace);
+	    parse_and_send_attribute (p, new_namespace, false);
 	  else
 	    parse_doit (p, true);
 	}
 
-      _gst_msg_sendf (NULL, "%v %o current: %o",
-		      _gst_namespace_class, old_namespace);
-      
-      _gst_unregister_oop (old_namespace);
+      set_compilation_namespace (old_namespace);
+      INC_RESTORE_POINTER (incPtr);
       return true;
     } 
 
@@ -685,11 +874,11 @@ parse_namespace_definition (gst_parser *p, tree_node first_stmt)
 static mst_Boolean
 parse_class_definition (gst_parser *p, OOP classOOP, mst_Boolean extend)
 {
-  int t1, t2, t3;
   mst_Boolean add_inst_vars = extend;
 
   for (;;)
     {
+      int t1, t2, t3;
       if (_gst_had_error)
 	break;
 
@@ -697,14 +886,11 @@ parse_class_definition (gst_parser *p, OOP classOOP, mst_Boolean extend)
       if (token (p, 0) == ']' || token (p, 0) == EOF)
 	break;
 
-      lex_lookahead (p, 3);
 #if 0
       print_tokens (p);      
 #endif
       
       t1 = token (p, 0);
-      t2 = token (p, 1);
-      t3 = token (p, 2);
 
       switch (t1) 
 	{	
@@ -722,6 +908,8 @@ parse_class_definition (gst_parser *p, OOP classOOP, mst_Boolean extend)
 	  continue;
 	  
 	case '<':
+	  lex_lookahead (p, 2);
+	  t2 = token (p, 1);
 	  if (t2 == IDENTIFIER) 
 	    {
 #if 0
@@ -735,12 +923,14 @@ parse_class_definition (gst_parser *p, OOP classOOP, mst_Boolean extend)
 	    }
 	  else if (t2 == KEYWORD) 
 	    {
-	      parse_and_send_attribute (p, classOOP);
+	      parse_and_send_attribute (p, classOOP, true);
 	      continue;
 	    }
 	  break;
    
 	case IDENTIFIER:
+	  lex_lookahead (p, 2);
+	  t2 = token (p, 1);
 	  if (t2 == ASSIGNMENT)
 	    {
 #if 0
@@ -766,13 +956,12 @@ parse_class_definition (gst_parser *p, OOP classOOP, mst_Boolean extend)
 		  class = (gst_class) OOP_TO_OBJ (the_class);
 		  class->classVariables = class_var_dict;
 		}
-	      
+
 	      stmt = parse_required_expression (p);
 	      if (!_gst_had_error)
 		{
 	          stmt = _gst_make_statement_list (&stmt->location, stmt);
-	          result = _gst_execute_statements (NULL, stmt, UNDECLARED_NONE,
-						    true);
+	          result = execute_doit (p, NULL, stmt, the_class, false, true);
 
 	          if (result)
 		    DICTIONARY_AT_PUT (class_var_dict, name, result);
@@ -814,6 +1003,8 @@ parse_class_definition (gst_parser *p, OOP classOOP, mst_Boolean extend)
 	    }
 	  else if (t2 == IDENTIFIER)
 	    {
+	      lex_lookahead (p, 3);
+	      t3 = token (p, 2);
 	      if (t3 == BINOP) 
 		{
 #if 0
@@ -853,6 +1044,8 @@ parse_class_definition (gst_parser *p, OOP classOOP, mst_Boolean extend)
 	  break;
 	  
 	case '|':
+	  lex_lookahead (p, 2);
+	  t2 = token (p, 1);
 	  if (t2  == '|') 
 	    {
 #if 0
@@ -863,6 +1056,8 @@ parse_class_definition (gst_parser *p, OOP classOOP, mst_Boolean extend)
 	    }
 	  else if (t2 == IDENTIFIER) 
 	    {
+	      lex_lookahead (p, 3);
+	      t3 = token (p, 2);
 	      if (t3 == IDENTIFIER || t3 == '|') 
 		{
 #if 0
@@ -904,7 +1099,7 @@ parse_scoped_method (gst_parser *p, OOP classOOP)
   mst_Boolean class_method = false;
 
   class_node = parse_variable_primary (p);
-  class = parse_class (class_node);
+  class = parse_class (p, class_node);
 
   if (OOP_CLASS (classOOP) == _gst_metaclass_class)
     classInstanceOOP = METACLASS_INSTANCE (classOOP);
@@ -961,10 +1156,11 @@ parse_scoped_method (gst_parser *p, OOP classOOP)
 }
 
 static OOP
-parse_class (tree_node list) 
+parse_class (gst_parser *p,
+	     tree_node list) 
 {
   const char* name;
-  OOP currentOOP = _gst_current_namespace;
+  OOP currentOOP = p->current_namespace;
   tree_node next; 
 
   if (strcmp (list->v_list.name, "nil") == 0)
@@ -1010,12 +1206,12 @@ parse_class (tree_node list)
 }
 
 static OOP
-parse_namespace (tree_node list) 
+parse_namespace (gst_parser *p, tree_node list) 
 {
   OOP name, new_namespace, current_namespace;
   const char *namespc;
 	
-  current_namespace = _gst_current_namespace;
+  current_namespace = p->current_namespace;
   while (list->v_list.next != NULL)
     {
       name = _gst_intern_string (list->v_list.name);	
@@ -1103,9 +1299,15 @@ parse_instance_variables (gst_parser *p, OOP classOOP, mst_Boolean extend)
 static void
 parse_method_list (gst_parser *p)
 {
-  while (token (p, 0) != '!')
-    parse_method (p, '!');
+  jmp_buf old_recover;
+  memcpy (old_recover, p->recover, sizeof (p->recover));
+  if (setjmp (p->recover) == 0)
+    {
+      while (token (p, 0) != '!')
+        parse_method (p, '!');
+    }
 
+  memcpy (p->recover, old_recover, sizeof (p->recover));
   _gst_skip_compilation = false;
   _gst_reset_compilation_category ();
   p->state = PARSE_DOIT;
@@ -1145,17 +1347,19 @@ parse_method (gst_parser *p, int at_end)
     current_pos.file_offset++;
 
   method = _gst_make_method (&pat->location, &current_pos,
-			     pat, temps, attrs, stmts,
+			     pat, temps, attrs, stmts, NULL,
+			     _gst_current_parser->currentClass,
+			     _gst_current_parser->currentCategory,
 			     at_end != ']');
 
   if (!_gst_had_error && !_gst_skip_compilation)
     {
-      enum undeclared_strategy oldUndeclared;
-      oldUndeclared = _gst_set_undeclared (UNDECLARED_GLOBALS);
-      _gst_compile_method (method, false, true);
-      _gst_set_undeclared (oldUndeclared);
+      _gst_current_parser->lastMethodOOP =
+        _gst_compile_method (method, false, true, false);
+      INC_ADD_OOP (_gst_current_parser->lastMethodOOP);
     }
 
+  assert (p->la_size <= 1);
   _gst_free_tree ();
   _gst_had_error = false;
   if (at_end != EOF)
@@ -1264,11 +1468,35 @@ parse_attributes (gst_parser *p, tree_node prev_attrs)
 static tree_node
 parse_attribute (gst_parser *p)
 {
-  tree_node attr;
+  tree_node message, attr, constant;
+  OOP attributeOOP, selectorOOP, argsOOP;
+  char *sel;
+  YYLTYPE location = *loc (p, 0);
+
   lex_skip_mandatory (p, '<');
-  lex_must_be (p, KEYWORD);
-  attr = parse_keyword_list (p, EXPR_BINOP);
-  attr = _gst_make_attribute_list (&attr->location, attr);
+  if (token (p, 0) == IDENTIFIER)
+    {
+      sel = val(p, 0)->sval;
+      lex (p);
+      selectorOOP = _gst_intern_string (sel);
+      new_instance_with (_gst_array_class, 0, &argsOOP);
+      MAKE_OOP_READONLY (selectorOOP, true);
+      MAKE_OOP_READONLY (argsOOP, true);
+      message = _gst_make_unary_expr (&location, NULL, sel);
+      attributeOOP = _gst_message_new_args (selectorOOP, argsOOP);
+    }
+  else
+    {
+      lex_must_be (p, KEYWORD);
+      message = parse_keyword_list (p, EXPR_BINOP);
+
+      /* First convert the TREE_KEYWORD_EXPR into a Message object, then
+	 into a TREE_CONST_EXPR, and finally embed this one into a
+	 TREE_ATTRIBUTE_LIST.  */
+      attributeOOP = make_attribute (p, _gst_nil_oop, message);
+    }
+  constant = _gst_make_oop_constant (&message->location, attributeOOP);
+  attr = _gst_make_attribute_list (&constant->location, constant);
   lex_skip_mandatory (p, '>');
   return attr;
 }
@@ -1720,7 +1948,6 @@ parse_compile_time_constant (gst_parser *p)
 {
   tree_node temps, statements;
   YYLTYPE location = *loc(p,0);
-  OOP result = NULL;
 
   assert (token (p, 0) == '#');
   lex (p);
@@ -1729,10 +1956,14 @@ parse_compile_time_constant (gst_parser *p)
   statements = parse_statements (p, NULL, true);
   lex_skip_mandatory (p, ')');
 
-  if (statements && !_gst_had_error)
-    result = _gst_execute_statements (temps, statements, UNDECLARED_NONE, true);
+  if (!statements || _gst_had_error)
+    return _gst_make_oop_constant (&location, _gst_nil_oop);
 
-  return _gst_make_oop_constant (&location, result ? result : _gst_nil_oop); 
+  return _gst_make_method (&location, loc(p, 0),
+                           NULL, temps, NULL, statements, NULL,
+                           _gst_current_parser->currentClass,
+                           _gst_nil_oop,
+                           false);
 }
 
 
