@@ -78,6 +78,10 @@
 #include <sys/time.h>
 #include <time.h>
 
+#ifdef HAVE_CRT_EXTERNS_H
+#include <crt_externs.h>
+#endif
+
 #ifdef HAVE_SYS_RESOURCE_H
 #include <sys/resource.h>
 #endif
@@ -199,9 +203,10 @@
   DO_PREFETCH ((x), 0, (k));
 
 /* Synchronization primitives.  */
-#if !defined HAVE_SYNC_BUILTINS && defined __GNUC__
-#define __sync_synchronize()	__asm__ ("" : : : "memory")
-#endif
+#define __sync_swap(ptr, val) \
+  ({ __typeof__ (*(ptr)) _x; \
+   do _x = *(ptr); while (!__sync_bool_compare_and_swap ((ptr), (_x), (val))); \
+   _x; })
 
 /* Kill a warning when using GNU C.  Note that this allows using
    break or continue inside a macro, unlike do...while(0) */
@@ -214,15 +219,6 @@
 #endif
 
 
-/* ENABLE_SECURITY enables security checks in the primitives as well as
-   special marking of untrusted objects.  Note that the code in the
-   class library to perform the security checks will be present
-   notwithstanding the setting of this flag, but they will be disabled
-   because the corresponding primitives will be made non-working.  We
-   define it here with no configure-time options because it causes
-   testsuite failures.  */
-#define ENABLE_SECURITY
-
 /* OPTIMIZE disables many checks, including consistency checks at GC
    time and bounds checking on instance variable accesses (not on #at:
    and #at:put: which would violate language semantics).  It can a)
@@ -261,24 +257,10 @@ extern char *strdup ();
    from the bottom.
 
    bit 0-3: reserved for distinguishing byte objects and saving their size.
-   bit 4-11: non-volatile bits (special kinds of objects).
-   bit 12-23: volatile bits (GC/JIT-related).
-   bit 24-30: reserved for counting things.
+   bit 4-14: non-volatile bits (special kinds of objects).  Used up to 11.
+   bit 15-30: volatile bits (GC/JIT-related).  Used up to 23.
    bit 31: unused to avoid signedness mess. */
 enum {
-  /* Place to save various OOP counts (how many fields we have marked
-     in the object, how many pointer instance variables there are, 
-     etc.).  Here is a distribution of frequencies in a standard image:
-       2 to   31 words      24798 objects (96.10%)
-      32 to   63 words        816 objects ( 3.16%)
-      64 to   95 words         82 objects ( 0.32%)
-      96 to  127 words         54 objects ( 0.21%)
-     128 or more words         54 objects ( 0.21%)
-
-    which I hope justifies the choice :-) */
-  F_COUNT = (int) 0x7F000000U,
-  F_COUNT_SHIFT = 24,
-
   /* Set if the object is reachable, during the mark phases of oldspace
      garbage collection.  */
   F_REACHABLE = 0x800000U,
@@ -316,7 +298,7 @@ enum {
 
   /* The grouping of all the flags which are not valid across image
      saves and loads.  */
-  F_RUNTIME = 0xFF8000U,
+  F_RUNTIME = 0x7FFF8000U,
 
   /* Set if the references to the instance variables of the object
      are weak.  */
@@ -339,9 +321,8 @@ enum {
   /* Set together with F_OLD for objects that live in fixedspace.  */
   F_FIXED = 0x200U,
 
-  /* Set for untrusted classes, instances of untrusted classes,
-     and contexts whose receiver is untrusted.  */
-  F_UNTRUSTED = 0x400U,
+  /* Free bit. */
+  F_FREE = 0x400U,
 
   /* Set for objects that were loaded from the image.  We never
      garbage collect their contents, only the OOPs.  */
@@ -361,37 +342,21 @@ enum {
 
 /* Answer whether a method, OOP, has already been verified. */
 #define IS_OOP_VERIFIED(oop) \
-  ((oop)->flags & F_VERIFIED)
+  (((oop)->flags & F_VERIFIED) != 0)
 
 /* Answer whether an object, OOP, is weak.  */
 #define IS_OOP_WEAK(oop) \
-  ((oop)->flags & F_WEAK)
+  (((oop)->flags & F_WEAK) != 0)
 
 /* Answer whether an object, OOP, is readonly.  */
 #define IS_OOP_READONLY(oop) \
-  (IS_INT ((oop)) ? F_READONLY : (oop)->flags & F_READONLY)
+  (IS_INT ((oop)) || ((oop)->flags & F_READONLY))
 
 /* Set whether an object, OOP, is readonly or readwrite.  */
 #define MAKE_OOP_READONLY(oop, ro) \
   (((oop)->flags &= ~F_READONLY), \
    ((oop)->flags |= (ro) ? F_READONLY : 0))
 
-#ifdef ENABLE_SECURITY
-
-/* Answer whether an object, OOP, is untrusted.  */
-#define IS_OOP_UNTRUSTED(oop) \
-  (IS_INT ((oop)) ? 0 : ((oop)->flags & F_UNTRUSTED))
-
-/* Set whether an object, OOP, is trusted or untrusted.  */
-#define MAKE_OOP_UNTRUSTED(oop, untr) \
-  (((oop)->flags &= ~F_UNTRUSTED), \
-   ((oop)->flags |= (untr) ? F_UNTRUSTED : 0))
-
-#else
-#define IS_OOP_UNTRUSTED(oop) (false)
-#define MAKE_OOP_UNTRUSTED(oop, untr) ((void)0)
-#endif
-
 /* Set whether an object, OOP, has ephemeron semantics.  */
 #define MAKE_OOP_EPHEMERON(oop) \
   (oop)->flags |= F_EPHEMERON;
@@ -518,8 +483,10 @@ extern OOP _gst_nil_oop
 #ifdef __GNUC__
 #define no_opt(x)	({ __typeof__ ((x)) _result; \
 			 asm ("" : "=r" (_result) : "0" ((x))); _result; })
+#define barrier()       asm ("")
 #else
 #define no_opt(x)	(x)
+#define barrier()
 #endif
 
 /* integer conversions and some information on SmallIntegers.  */
@@ -590,25 +557,24 @@ extern OOP _gst_nil_oop
 #include "strspell.h"
 #include "alloc.h"
 #include "md-config.h"
-#include "memzero.h"
 #include "avltrees.h"
 #include "rbtrees.h"
 
-#include "tree.h"
 #include "files.h"
-#include "input.h"
 #include "callin.h"
 #include "cint.h"
 #include "dict.h"
-#include "events.h"
 #include "heap.h"
 #include "lex.h"
+#include "tree.h"
+#include "sym.h"
 #include "gst-parse.h"
+#include "input.h"
 #include "oop.h"
 #include "byte.h"
-#include "sym.h"
 #include "comp.h"
 #include "interp.h"
+#include "events.h"
 #include "opt.h"
 #include "save.h"
 #include "str.h"
@@ -616,7 +582,6 @@ extern OOP _gst_nil_oop
 #include "xlat.h"
 #include "mpz.h"
 #include "print.h"
-#include "security.h"
 #include "real.h"
 #include "sockets.h"
 
