@@ -7,7 +7,8 @@
 
 /***********************************************************************
  *
- * Copyright 2001, 2002, 2003, 2005, 2006 Free Software Foundation, Inc.
+ * Copyright 2001, 2002, 2003, 2005, 2006, 2008, 2009
+ * Free Software Foundation, Inc.
  * Written by Paolo Bonzini.
  *
  * This file is part of GNU Smalltalk.
@@ -60,7 +61,7 @@
 typedef struct gst_file_segment
 {
   OBJ_HEADER;
-  OOP fileName;
+  OOP fileOOP;
   OOP startPos;
   OOP length;
 }
@@ -104,7 +105,7 @@ typedef struct input_stream
   int column;
   const char *prompt;
 
-  OOP fileNameOOP;		/* the full path name for file */
+  OOP fileOOP;			/* the object stored in FileSegments */
   const char *fileName;
   off_t fileOffset;
 
@@ -126,6 +127,10 @@ typedef struct input_stream
 /* The internal interface used by _gst_next_char.  */
 static int my_getc (input_stream stream);
 
+/* Return the File object or a file name for the topmost stream in the stack
+   if it is of type STREAM_FILE; nil otherwise.  */
+static OOP get_cur_file (void);
+
 /* Print a line indicator in front of an error message.  */
 static void line_stamp (int line);
 
@@ -141,8 +146,6 @@ static input_stream in_stream = NULL;
    POLLHUP) and then perform a read system call.  */
 static int poll_and_read (int fd, char *buf, int n);
 
-static int change_str = -1;
-
 
 /* If true, readline is disabled.  */
 mst_Boolean _gst_no_tty = false;
@@ -214,9 +217,7 @@ _gst_pop_stream (mst_Boolean closeIt)
 
   stream = in_stream;
   in_stream = in_stream->prevStream;
-
-  if (!IS_NIL (stream->fileNameOOP))
-    _gst_unregister_oop (stream->fileNameOOP);
+  _gst_unregister_oop (stream->fileOOP);
 
   switch (stream->type)
     {
@@ -336,7 +337,7 @@ push_new_stream (stream_type type)
   newStream->type = type;
   newStream->fileName = NULL;
   newStream->prompt = NULL;
-  newStream->fileNameOOP = _gst_nil_oop;
+  newStream->fileOOP = _gst_nil_oop;
   newStream->prevStream = in_stream;
   in_stream = newStream;
   return (newStream);
@@ -345,19 +346,19 @@ push_new_stream (stream_type type)
 
 void
 _gst_set_stream_info (int line,
+		      OOP fileOOP,
 		      OOP fileNameOOP,
 		      int fileOffset)
 {
   in_stream->line = line;
   in_stream->column = 0;
-  if (!IS_NIL (fileNameOOP))
-    {
-      in_stream->fileName = _gst_to_cstring (fileNameOOP);
-      _gst_register_oop (in_stream->fileNameOOP);
-    }
 
-  in_stream->fileNameOOP = fileNameOOP;
+  _gst_register_oop (fileOOP);
+  in_stream->fileOOP = fileOOP;
   in_stream->fileOffset = fileOffset;
+
+  if (!IS_NIL (fileNameOOP))
+    in_stream->fileName = _gst_to_cstring (fileNameOOP);
 }
  
 void
@@ -404,7 +405,7 @@ my_getc (input_stream stream)
       if (stream->st_oop.ptr == stream->st_oop.end)
 	{
 	  char *buf;
-	  _gst_msg_sendf(&buf, "%s %o nextHunk", stream->st_oop.oop);
+	  _gst_msg_sendf(&buf, "%s %o nextAvailable: %i", stream->st_oop.oop, 1024);
 	  if (!buf || !*buf)
 	    return EOF;
 
@@ -487,26 +488,32 @@ _gst_get_source_string (off_t startPos, off_t endPos)
   if (!in_stream)
     return (_gst_nil_oop);
 
+  /* FIXME: check isPipe too? */
   if (startPos != -1 && !_gst_get_cur_stream_prompt ())
     {
-      OOP fileName;
+      OOP fileOOP;
       gst_file_segment fileSegment;
       inc_ptr incPtr;
 
       incPtr = INC_SAVE_POINTER ();
-      fileName = _gst_get_cur_file_name ();
-      INC_ADD_OOP (fileName);
+      fileOOP = get_cur_file ();
 
-      fileSegment = (gst_file_segment) new_instance (_gst_file_segment_class,
-                                                     &result);
+      if (!IS_NIL (fileOOP))
+	{
+          INC_ADD_OOP (fileOOP);
+          fileSegment = (gst_file_segment) new_instance (_gst_file_segment_class,
+                                                         &result);
 
-      fileSegment->fileName = fileName;
-      fileSegment->startPos = from_c_int_64 (startPos);
-      fileSegment->length = from_c_int_64 (endPos - startPos);
+          fileSegment->fileOOP = fileOOP;
+          fileSegment->startPos = from_c_int_64 (startPos);
+          fileSegment->length = from_c_int_64 (endPos - startPos);
+
+          assert (to_c_int_64 (fileSegment->length) >= 0);
+          INC_RESTORE_POINTER (incPtr);
+          return (result);
+	}
 
-      assert (to_c_int_64 (fileSegment->length) >= 0);
       INC_RESTORE_POINTER (incPtr);
-      return (result);
     }
 
   switch (in_stream->type)
@@ -519,6 +526,7 @@ _gst_get_source_string (off_t startPos, off_t endPos)
     case STREAM_READLINE:
 #endif /* HAVE_READLINE */
     case STREAM_OOP:
+    case STREAM_FILE:
       p = in_stream->st_oop.buf;
       break;
 
@@ -548,15 +556,15 @@ _gst_get_source_string (off_t startPos, off_t endPos)
 }
 
 OOP
-_gst_get_cur_file_name (void)
+get_cur_file (void)
 {
-  char *fullFileName;
+  const char *fullFileName;
 
   if (!in_stream)
     return _gst_nil_oop;
 
-  if (!IS_NIL (in_stream->fileNameOOP))
-    return in_stream->fileNameOOP;
+  if (!IS_NIL (in_stream->fileOOP))
+    return in_stream->fileOOP;
 
   if (in_stream->type != STREAM_FILE)
     return (_gst_nil_oop);
@@ -567,9 +575,9 @@ _gst_get_cur_file_name (void)
     fullFileName =
       _gst_get_full_file_name (in_stream->fileName);
 
-  in_stream->fileNameOOP = _gst_string_new (fullFileName);
-  _gst_register_oop (in_stream->fileNameOOP);
-  return (in_stream->fileNameOOP);
+  in_stream->fileOOP = _gst_string_new (fullFileName);
+  _gst_register_oop (in_stream->fileOOP);
+  return (in_stream->fileOOP);
 }
 
 
@@ -700,14 +708,7 @@ line_stamp (int line)
   if (_gst_report_errors)
     {
       if (in_stream)
-	{
-	  if (in_stream->fileName)
-	    fprintf (stderr, "%s:", in_stream->fileName);
-	  else if (!IS_NIL (in_stream->fileNameOOP))
-	    fprintf (stderr, "%#O:", in_stream->fileNameOOP);
-
-	  fprintf (stderr, "%d: ", line);
-	}
+	fprintf (stderr, "%s:%d: ", in_stream->fileName, line);
       else
 	fprintf (stderr, "gst: ");
     }
@@ -747,10 +748,6 @@ _gst_next_char (void)
   else
     {
       ic = my_getc (in_stream);
-
-      if (change_str != -1)
-	_gst_full_write (change_str, &ic, 1);
-
       if (ic == '\n')
 	{			/* a new line that was not pushed back */
 	  in_stream->line++;
@@ -789,21 +786,9 @@ int
 poll_and_read (int fd, char *buf, int n)
 {
   int result;
-  struct pollfd pfd;
 
-  pfd.fd = fd;
-  pfd.events = POLLIN;
-  pfd.revents = 0;
-
-  do
-    {
-      errno = 0;
-      result = poll (&pfd, 1, -1); /* Infinite wait */
-    }
-  while ((result == 0 && (pfd.revents & POLLHUP) == 0)
-	 || ((result == -1) && (errno == EINTR)));
-
-  if (pfd.revents & POLLIN)
+  _gst_wait_for_input (fd);
+  if (_gst_sync_file_polling (fd, 0))
     {
       do
 	{
@@ -830,7 +815,7 @@ _gst_process_stdin (const char *prompt)
   _gst_push_stdin_string ();
   if (isatty (0))
     in_stream->prompt = prompt;
-  _gst_parse_stream (false);
+  _gst_parse_stream (NULL);
   _gst_pop_stream (true);
   _gst_non_interactive = true;
 }
@@ -838,7 +823,6 @@ _gst_process_stdin (const char *prompt)
 mst_Boolean
 _gst_process_file (const char *fileName, enum gst_file_dir dir)
 {
-  enum undeclared_strategy old;
   int fd;
   char *f;
 
@@ -856,11 +840,9 @@ _gst_process_file (const char *fileName, enum gst_file_dir dir)
       if (_gst_verbosity == 3)
 	printf ("Processing %s\n", f);
 
-      old = _gst_set_undeclared (UNDECLARED_GLOBALS);
       _gst_push_unix_file (fd, f);
-      _gst_parse_stream (false);
+      _gst_parse_stream (NULL);
       _gst_pop_stream (true);
-      _gst_set_undeclared (old);
       errno = 0;
     }
 
@@ -960,8 +942,8 @@ _gst_add_symbol_completion (const char *str,
 	}
     }
 
-  /* We enter globals in the table, too */
-  if (base == str && !islower (*str))
+  /* We enter all unary and binary symbols in the table, too */
+  if (base == str)
     add_completion (base, p - base);
 }
 
