@@ -1,6 +1,6 @@
 /******************************** -*- C -*- ****************************
  *
- *	The Smalltalk Virtual Machine in itself.
+ *	The Smalltalk Virtual Machine itself.
  *
  *	This, together with oop.c, is the `bridge' between Smalltalk and
  *	the underlying machine
@@ -10,7 +10,7 @@
 
 /***********************************************************************
  *
- * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2006,2007,2008
+ * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2006,2007,2008,2009
  * Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
@@ -57,8 +57,6 @@
 #include "gstpriv.h"
 #include "lock.h"
 
-
-
 /* The local regs concept hopes, by caching the values of IP and SP in
    local register variables, to increase performance.  You only need
    to export the variables when calling out to routines that might
@@ -115,13 +113,6 @@
    to speed up these messages for Arrays, Strings, and ByteArrays.  */
 #define METHOD_CACHE_SIZE		(1 << 11)
 
-typedef struct async_queue_entry
-{
-  void (*func) (OOP);
-  OOP data;
-}
-async_queue_entry;
-
 typedef struct interp_jmp_buf
 {
   jmp_buf jmpBuf;
@@ -171,6 +162,12 @@ unsigned long _gst_cache_misses = 0;
 /* The number of cache lookups - either hits or misses */
 unsigned long _gst_sample_counter = 0;
 
+/* The OOP for an IdentityDictionary that stores the raw profile.  */
+OOP _gst_raw_profile = NULL;
+
+/* A bytecode counter value used while profiling. */
+unsigned long _gst_saved_bytecode_counter = 0;
+
 #ifdef ENABLE_JIT_TRANSLATION
 #define method_base		0
 char *native_ip = NULL;
@@ -243,14 +240,10 @@ static int class_cache_prim;
 #endif
 
 /* Queue for async (outside the interpreter) semaphore signals */
-gl_lock_define (static, async_queue_lock);
 static mst_Boolean async_queue_enabled = true;
-static int async_queue_index;
-static int async_queue_size;
-static async_queue_entry *queued_async_signals;
-
-static int async_queue_index_sig;
-static async_queue_entry queued_async_signals_sig[NSIG];
+static async_queue_entry queued_async_signals_tail;
+static async_queue_entry *queued_async_signals = &queued_async_signals_tail;
+static async_queue_entry *queued_async_signals_sig = &queued_async_signals_tail;
 
 /* When not NULL, this causes the byte code interpreter to immediately
    send the message whose selector is here to the current stack
@@ -291,6 +284,19 @@ static inline mst_Boolean cached_index_oop_put_primitive (OOP rec,
 							  OOP val,
 							  intptr_t spec);
 
+/* This functions accepts an OOP for a Semaphore object and puts the
+   PROCESSOOP to sleep, unless the semaphore has excess signals
+   on it.  */
+static void sync_wait_process (OOP semaphoreOOP,
+			       OOP processOOP);
+
+/* Empty the queue of asynchronous calls.  */
+static void empty_async_queue (void);
+
+/* Try to find another process with higher or same priority as the
+   active one.  Return whether there is one.  */
+static mst_Boolean would_reschedule_process (void);
+
 /* Locates in the ProcessorScheduler's process lists and returns the
    highest priority process different from the current process.  */
 static OOP highest_priority_process (void);
@@ -325,13 +331,22 @@ static void set_preemption_timer (void);
 
 /* Same as _gst_parse_stream, but creating a reentrancy_jmpbuf.  Returns
    true if interrupted. */
-static mst_Boolean parse_stream_with_protection (mst_Boolean method);
+static mst_Boolean parse_stream_with_protection (OOP currentNamespace);
+
+/* Same as _gst_parse_method_from_stream, but creating a reentrancy_jmpbuf.
+   Returns true if interrupted, pushes the last compiled method on the
+   stack. */
+static mst_Boolean parse_method_from_stream_with_protection (OOP currentClass,
+							     OOP currentCategory);
 
 /* Put the given process to sleep by rotating the list of processes for
    PROCESSOOP's priority (i.e. it was the head of the list and becomes
    the tail).  */
 static void sleep_process (OOP processOOP);
 
+/* Yield control from the active process.  */
+static void active_process_yield (void);
+
 /* Sets flags so that the interpreter switches to PROCESSOOP at the
    next sequence point.  Unless PROCESSOOP is already active, in which
    case nothing happens, the process is made the head of the list of
@@ -455,7 +470,7 @@ static inline void prepare_context (gst_context_part context,
 
 /* Return from the current context and restore the virtual machine's
    status (ip, sp, _gst_this_method, _gst_self, ...).  */
-static void unwind_context (void);
+static void __attribute__ ((__always_inline__)) unwind_context (void);
 
 /* Check whether it is true that sending SENDSELECTOR to RECEIVER
    accepts NUMARGS arguments.  Note that the RECEIVER is only used to
@@ -567,21 +582,6 @@ static void * const *dispatch_vec;
 #define PARENT_CONTEXT(contextOOP) \
   ( ((gst_method_context) OOP_TO_OBJ (contextOOP)) ->parentContext)
 
-/* Set whether the old context was a trusted one.  Untrusted contexts
-   are those whose receiver or sender is untrusted.  */
-#define UPDATE_CONTEXT_TRUSTFULNESS(contextOOP, parentContextOOP) \
-  MAKE_OOP_UNTRUSTED (contextOOP, \
-    IS_OOP_UNTRUSTED (_gst_self) | \
-    IS_OOP_UNTRUSTED (parentContextOOP));
-
-/* Set whether the current context is an untrusted one.  Untrusted contexts
-   are those whose receiver or sender is untrusted.  */
-#define IS_THIS_CONTEXT_UNTRUSTED() \
-  (UPDATE_CONTEXT_TRUSTFULNESS(_gst_this_context_oop, \
-			       PARENT_CONTEXT (_gst_this_context_oop)) \
-     & F_UNTRUSTED)
-
-
 /* Context management
  
    The contexts make up a linked list.  Their structure is:
@@ -754,8 +754,6 @@ empty_context_stack (void)
   context->spOffset = FROM_INT (sp - context->contextStack);
   context->ipOffset = FROM_INT (ip - method_base);
 
-  UPDATE_CONTEXT_TRUSTFULNESS (_gst_this_context_oop, context->parentContext);
-
   /* Even if the JIT is active, the current context might have no
      attached native_ip -- in fact it has one only if we are being
      called from activate_new_context -- so we have to `invent'
@@ -769,8 +767,6 @@ empty_context_stack (void)
 void
 alloc_new_chunk (void)
 {
-  gst_method_context newContext;
-
   if UNCOMMON (++chunk >= &chunks[MAX_CHUNKS_IN_MEMORY])
     {
       /* No more chunks available - GC */
@@ -780,20 +776,13 @@ alloc_new_chunk (void)
 
   empty_context_stack ();
 
-  newContext = (gst_method_context) * chunk;
-  if UNCOMMON (!newContext)
-    {
-      /* Allocate memory only the first time we're using the chunk.
-         _gst_empty_context_pool resets the status but doesn't free
-         the memory.  */
-      cur_chunk_begin = *chunk = (gst_context_part)
-        xcalloc (1, SIZE_TO_BYTES (CHUNK_SIZE));
-
-      newContext = (gst_method_context) cur_chunk_begin;
-    }
-  else
-    cur_chunk_begin = *chunk;
+  /* Allocate memory only the first time we're using the chunk.
+     _gst_empty_context_pool resets the status but doesn't free
+     the memory.  */
+  if UNCOMMON (!*chunk)
+    *chunk = (gst_context_part) xcalloc (1, SIZE_TO_BYTES (CHUNK_SIZE));
 
+  cur_chunk_begin = *chunk;
   cur_chunk_end = (gst_context_part) (
     ((char *) cur_chunk_begin) + SIZE_TO_BYTES(CHUNK_SIZE));
 }
@@ -857,7 +846,6 @@ activate_new_context (int size,
     FROM_INT ((sp - thisContext->contextStack) - sendArgs);
   thisContext->ipOffset = FROM_INT (ip - method_base);
 
-  UPDATE_CONTEXT_TRUSTFULNESS (_gst_this_context_oop, thisContext->parentContext);
   _gst_this_context_oop = oop;
 
   return (newContext);
@@ -884,25 +872,44 @@ prepare_context (gst_context_part context,
 		 int args,
 		 int temps)
 {
-  REGISTER (1, OOP * mySP);
-  _gst_temporaries = mySP = context->contextStack;
+  REGISTER (1, OOP *stackBase);
+  _gst_temporaries = stackBase = context->contextStack;
   if (args)
     {
-      REGISTER (2, int num);
-      REGISTER (3, OOP * src);
-      num = args;
-      src = &sp[1 - num];
-
-#define UNROLL_OP(n) mySP[n] = src[n]
-#define UNROLL_ADV(n) mySP += n, src += n
-      UNROLL_BY_8 (num);
-#undef UNROLL_OP
-#undef UNROLL_ADV
-
-      mySP += num;
+      REGISTER (2, OOP * src);
+      src = &sp[1 - args];
+      stackBase[0] = src[0];
+      if (args > 1)
+        {
+          stackBase[1] = src[1];
+          if (args > 2)
+            {
+              stackBase[2] = src[2];
+              if (args > 3)
+                memcpy (&stackBase[3], &src[3], (args - 3) * sizeof (OOP));
+            }
+        }
+      stackBase += args;
     }
-  sp = mySP + temps - 1;
-  nil_fill (mySP, temps);
+  if (temps)
+    {
+      REGISTER (2, OOP src);
+      src = _gst_nil_oop;
+      stackBase[0] = src;
+      if (temps > 1)
+        {
+          stackBase[1] = src;
+          if (temps > 2)
+            {
+              int n = 2;
+              do
+                stackBase[n] = src;
+              while UNCOMMON (n++ < temps);
+            }
+        }
+      stackBase += temps;
+    }
+  sp = stackBase - 1;
 }
 
 mst_Boolean
@@ -913,6 +920,8 @@ lookup_method (OOP sendSelector,
 {
   inc_ptr inc;
   OOP argsArrayOOP;
+  gst_message_lookup messageLookup;
+  OOP messageLookupOOP;
 
   if (_gst_find_method (method_class, sendSelector, methodData))
     return (true);
@@ -920,8 +929,16 @@ lookup_method (OOP sendSelector,
   inc = INC_SAVE_POINTER ();
   argsArrayOOP = create_args_array (sendArgs);
   INC_ADD_OOP (argsArrayOOP);
-  PUSH_OOP (_gst_message_new_args (sendSelector, argsArrayOOP));
+
+  messageLookup = (gst_message_lookup) new_instance (_gst_message_lookup_class,
+					             &messageLookupOOP);
+
+  messageLookup->selector = sendSelector;
+  messageLookup->args = argsArrayOOP;
+  messageLookup->startingClass = method_class;
+  PUSH_OOP (messageLookupOOP);
   INC_RESTORE_POINTER (inc);
+
   return (false);
 }
 
@@ -1003,36 +1020,30 @@ void
 unwind_context (void)
 {
   gst_method_context oldContext, newContext;
-  OOP oldContextOOP, newContextOOP;
-  int numLifoContexts;
+  OOP newContextOOP;
 
   newContextOOP = _gst_this_context_oop;
   newContext = (gst_method_context) OOP_TO_OBJ (newContextOOP);
-  numLifoContexts = free_lifo_context - lifo_contexts;
 
   do
     {
-      oldContextOOP = newContextOOP;
       oldContext = newContext;
 
       /* Descend in the chain...  */
       newContextOOP = oldContext->parentContext;
 
-      if COMMON (numLifoContexts > 0)
-	{
-	  dealloc_stack_context ((gst_context_part) oldContext);
-	  numLifoContexts--;
-	}
+      if COMMON (free_lifo_context > lifo_contexts)
+        dealloc_stack_context ((gst_context_part) oldContext);
 
-      else
-	/* This context cannot be deallocated in a LIFO way.  We must
-	   keep it around so that the blocks it created can reference
-	   arguments and temporaries in it. Method contexts, however,
-	   need to be marked as non-returnable so that attempts to
-	   return from them to an undefined place will lose; doing
-	   that for block contexts too, we skip a test and are also
-	   able to garbage collect more context objects.  */
-	oldContext->parentContext = _gst_nil_oop;
+      /* This context cannot be deallocated in a LIFO way.  We must
+         keep it around so that the blocks it created can reference
+         arguments and temporaries in it. Method contexts, however,
+         need to be marked as non-returnable so that attempts to
+         return from them to an undefined place will lose; doing
+         that for block contexts too, we skip a test and are also
+         able to garbage collect more context objects.  And doing
+         that for _all_ method contexts is more icache-friendly.  */
+      oldContext->parentContext = _gst_nil_oop;
 
       newContext = (gst_method_context) OOP_TO_OBJ (newContextOOP);
     }
@@ -1178,7 +1189,7 @@ unwind_to (OOP returnContextOOP)
 mst_Boolean
 disable_non_unwind_contexts (OOP returnContextOOP)
 {
-  OOP oldContextOOP, newContextOOP, *chain;
+  OOP newContextOOP, *chain;
   gst_method_context oldContext, newContext;
 
   newContextOOP = _gst_this_context_oop;
@@ -1187,7 +1198,6 @@ disable_non_unwind_contexts (OOP returnContextOOP)
 
   for (;;)
     {
-      oldContextOOP = newContextOOP;
       oldContext = newContext;
 
       /* Descend in the chain...  */
@@ -1225,7 +1235,6 @@ disable_non_unwind_contexts (OOP returnContextOOP)
   while UNCOMMON (CONTEXT_FLAGS (newContext)
                   == (MCF_IS_METHOD_CONTEXT | MCF_IS_DISABLED_CONTEXT))
     {
-      oldContextOOP = newContextOOP;
       oldContext = newContext;
 
       /* Descend in the chain...  */
@@ -1336,8 +1345,8 @@ resume_suspended_context (OOP oop)
 
   _gst_this_context_oop = oop;
   thisContext = (gst_method_context) OOP_TO_OBJ (oop);
-  SET_THIS_METHOD (thisContext->method, GET_CONTEXT_IP (thisContext));
   sp = thisContext->contextStack + TO_INT (thisContext->spOffset);
+  SET_THIS_METHOD (thisContext->method, GET_CONTEXT_IP (thisContext));
 
 #if ENABLE_JIT_TRANSLATION
   ip = TO_INT (thisContext->ipOffset);
@@ -1370,51 +1379,6 @@ get_scheduled_process (void)
   return (processor->activeProcess);
 }
 
-void
-add_first_link (OOP semaphoreOOP,
-		OOP processOOP)
-{
-  gst_semaphore sem;
-  gst_process process, lastProcess;
-  OOP lastProcessOOP;
-
-  process = (gst_process) OOP_TO_OBJ (processOOP);
-  if (!IS_NIL (process->myList))
-    {
-      sem = (gst_semaphore) OOP_TO_OBJ (process->myList);
-      if (sem->firstLink == processOOP)
-	{
-	  sem->firstLink = process->nextLink;
-	  if (sem->lastLink == processOOP)
-	    {
-	      /* It was the only process in the list */
-	      sem->lastLink = _gst_nil_oop;
-	    }
-	}
-      else if (sem->lastLink == processOOP)
-	{
-	  /* Find the new last link */
-	  lastProcessOOP = sem->firstLink;
-	  lastProcess = (gst_process) OOP_TO_OBJ (lastProcessOOP);
-	  while (lastProcess->nextLink != processOOP)
-	    {
-	      lastProcessOOP = lastProcess->nextLink;
-	      lastProcess = (gst_process) OOP_TO_OBJ (lastProcessOOP);
-	    }
-	  sem->lastLink = lastProcessOOP;
-	  lastProcess->nextLink = _gst_nil_oop;
-	}
-    }
-
-  sem = (gst_semaphore) OOP_TO_OBJ (semaphoreOOP);
-  process->myList = semaphoreOOP;
-  process->nextLink = sem->firstLink;
-
-  sem->firstLink = processOOP;
-  if (IS_NIL (sem->lastLink))
-    sem->lastLink = processOOP;
-}
-
 static void
 remove_process_from_list (OOP processOOP)
 {
@@ -1439,7 +1403,7 @@ remove_process_from_list (OOP processOOP)
         }
       else
         {
-          /* Find the new last link */
+          /* Find the new prev node */
           lastProcessOOP = sem->firstLink;
           lastProcess = (gst_process) OOP_TO_OBJ (lastProcessOOP);
           while (lastProcess->nextLink != processOOP)
@@ -1459,6 +1423,25 @@ remove_process_from_list (OOP processOOP)
   process->nextLink = _gst_nil_oop;
 }
 
+void
+add_first_link (OOP semaphoreOOP,
+		OOP processOOP)
+{
+  gst_semaphore sem;
+  gst_process process;
+
+  process = (gst_process) OOP_TO_OBJ (processOOP);
+  remove_process_from_list (processOOP);
+
+  sem = (gst_semaphore) OOP_TO_OBJ (semaphoreOOP);
+  process->myList = semaphoreOOP;
+  process->nextLink = sem->firstLink;
+
+  sem->firstLink = processOOP;
+  if (IS_NIL (sem->lastLink))
+    sem->lastLink = processOOP;
+}
+
 void
 suspend_process (OOP processOOP)
 {
@@ -1486,32 +1469,7 @@ add_last_link (OOP semaphoreOOP,
   OOP lastProcessOOP;
 
   process = (gst_process) OOP_TO_OBJ (processOOP);
-  if (!IS_NIL (process->myList))
-    {
-      sem = (gst_semaphore) OOP_TO_OBJ (process->myList);
-      if (sem->firstLink == processOOP)
-	{
-	  sem->firstLink = process->nextLink;
-	  if (sem->lastLink == processOOP)
-	    {
-	      /* It was the only process in the list */
-	      sem->lastLink = _gst_nil_oop;
-	    }
-	}
-      else if (sem->lastLink == processOOP)
-	{
-	  /* Find the new last link */
-	  lastProcessOOP = sem->firstLink;
-	  lastProcess = (gst_process) OOP_TO_OBJ (lastProcessOOP);
-	  while (lastProcess->nextLink != processOOP)
-	    {
-	      lastProcessOOP = lastProcess->nextLink;
-	      lastProcess = (gst_process) OOP_TO_OBJ (lastProcessOOP);
-	    }
-	  sem->lastLink = lastProcessOOP;
-	  lastProcess->nextLink = _gst_nil_oop;
-	}
-    }
+  remove_process_from_list (processOOP);
 
   sem = (gst_semaphore) OOP_TO_OBJ (semaphoreOOP);
   process->myList = semaphoreOOP;
@@ -1537,15 +1495,33 @@ is_empty (OOP processListOOP)
   return (IS_NIL (processList->firstLink));
 }
 
+/* TODO: this was taken from VMpr_Processor_yield.  Try to use
+   the macro ACTIVE_PROCESS_YIELD instead?  */
+
+void
+active_process_yield (void)
+{
+  OOP activeProcess = get_active_process ();
+  OOP newProcess = highest_priority_process();
+
+  if (is_process_ready (activeProcess))
+    sleep_process (activeProcess);	/* move to the end of the list */
+
+  activate_process (IS_NIL (newProcess) ? activeProcess : newProcess);
+}
+
 
 mst_Boolean
 _gst_sync_signal (OOP semaphoreOOP, mst_Boolean incr_if_empty)
 {
   gst_semaphore sem;
-  OOP freedOOP;
+  gst_process process;
+  gst_method_context suspendedContext;
+  OOP processOOP;
+  int spOffset;
 
   sem = (gst_semaphore) OOP_TO_OBJ (semaphoreOOP);
-  for (;;)
+  do
     {
       /* printf ("signal %O %O\n", semaphoreOOP, sem->firstLink); */
       if (is_empty (semaphoreOOP))
@@ -1555,87 +1531,174 @@ _gst_sync_signal (OOP semaphoreOOP, mst_Boolean incr_if_empty)
 	  return false;
 	}
 
-      freedOOP = remove_first_link (semaphoreOOP);
+      processOOP = remove_first_link (semaphoreOOP);
 
       /* If they terminated this process, well, try another */
-      if (resume_process (freedOOP, false))
-	return true;
     }
+  while (!resume_process (processOOP, false));
+
+  /* Put the semaphore at the stack top as a marker that the
+     wait was not interrupted.  This assumes that _gst_sync_wait
+     is only called from primitives.  */
+  process = (gst_process) OOP_TO_OBJ (processOOP);
+  suspendedContext = (gst_method_context) OOP_TO_OBJ (process->suspendedContext);
+  spOffset = TO_INT (suspendedContext->spOffset);
+  suspendedContext->contextStack[spOffset] = semaphoreOOP;
+  return true;
 }
 
-static void
-do_async_signal (OOP semaphoreOOP)
+void
+_gst_do_async_signal (OOP semaphoreOOP)
 {
   _gst_sync_signal (semaphoreOOP, true);
 }
 
+void
+_gst_do_async_signal_and_unregister (OOP semaphoreOOP)
+{
+  _gst_sync_signal (semaphoreOOP, true);
+  _gst_unregister_oop (semaphoreOOP);
+}
+
+/* Async-signal-safe version, does no allocation.  Using an atomic operation
+   is still the simplest choice, but on top of that we check that the entry
+   is not already in the list.  Also, the datum and next field are NULLed
+   automatically when the call is made.  */
+void
+_gst_async_call_internal (async_queue_entry *e)
+{
+  /* For async-signal safety, we need to check that the entry is not
+     already in the list.  Checking that atomically with CAS is the
+     simplest way.  */
+  do
+    if (__sync_val_compare_and_swap(&e->next, NULL, queued_async_signals_sig))
+      return;
+  while (!__sync_bool_compare_and_swap (&queued_async_signals_sig, e->next, e));
+  SET_EXCEPT_FLAG (true);
+}
+
 void
 _gst_async_call (void (*func) (OOP), OOP arg)
 {
-  if (_gst_signal_count)
-    {
-      /* Async-signal-safe version.  Because the discriminant is
-	 _gst_signal_count, can only be called from within libgst.la
-	 (in particular events.c).  Also, because of that we assume
-	 interrupts are already disabled.  */
+  /* Thread-safe version for the masses.  This lockless stack
+     is reversed in the interpreter loop to get FIFO behavior.  */
+  async_queue_entry *sig = xmalloc (sizeof (async_queue_entry));
+  sig->func = func;
+  sig->data = arg;
 
-      if (async_queue_index_sig == NSIG)
-        {
-	  static const char errmsg[] = "Asynchronous signal queue full!\n";
-	  write (2, errmsg, sizeof (errmsg) - 1);
-	  abort ();
-	}
-
-      queued_async_signals_sig[async_queue_index_sig].func = func;
-      queued_async_signals_sig[async_queue_index_sig++].data = arg;
-    }
-  else
-    {
-      /* Thread-safe version for the masses.  */
-
-      gl_lock_lock (async_queue_lock);
-      if (async_queue_index == async_queue_size)
-        {
-          async_queue_size *= 2;
-          queued_async_signals =
-	    realloc (queued_async_signals,
-		 sizeof (async_queue_entry) * async_queue_size);
-	}
-
-      queued_async_signals[async_queue_index].func = func;
-      queued_async_signals[async_queue_index++].data = arg;
-      gl_lock_unlock (async_queue_lock);
-    }
-  
+  do
+    sig->next = queued_async_signals;
+  while (!__sync_bool_compare_and_swap (&queued_async_signals,
+                                        sig->next, sig));
+  _gst_wakeup ();
   SET_EXCEPT_FLAG (true);
 }
 
+mst_Boolean
+_gst_have_pending_async_calls ()
+{
+  return (queued_async_signals != &queued_async_signals_tail
+          || queued_async_signals_sig != &queued_async_signals_tail);
+}
+
+void
+empty_async_queue ()
+{
+  async_queue_entry *sig, *sig_reversed;
+
+  /* Process a batch of asynchronous requests.  These are pushed
+     in LIFO order by _gst_async_call.  By reversing the list
+     in place before walking it, we get FIFO order.  */
+  sig = __sync_swap (&queued_async_signals, &queued_async_signals_tail);
+  sig_reversed = &queued_async_signals_tail;
+  while (sig != &queued_async_signals_tail)
+    {
+      async_queue_entry *next = sig->next;
+      sig->next = sig_reversed;
+      sig_reversed = sig;
+      sig = next;
+    }
+
+  sig = sig_reversed;
+  while (sig != &queued_async_signals_tail)
+    {
+      async_queue_entry *next = sig->next;
+      sig->func (sig->data);
+      free (sig);
+      sig = next;
+    }
+
+  /* For async-signal-safe processing, we need to avoid entering
+     the same item twice into the list.  So we use NEXT to mark
+     items that have been added...  */
+  sig = __sync_swap (&queued_async_signals_sig, &queued_async_signals_tail);
+  sig_reversed = &queued_async_signals_tail;
+  while (sig != &queued_async_signals_tail)
+    {
+      async_queue_entry *next = sig->next;
+      sig->next = sig_reversed;
+      sig_reversed = sig;
+      sig = next;
+    }
+
+  sig = sig_reversed;
+  while (sig != &queued_async_signals_tail)
+    {
+      async_queue_entry *next = sig->next;
+      void (*func) (OOP) = sig->func;
+      OOP data = sig->data;
+      barrier ();
+
+      sig->data = NULL;
+      barrier ();
+
+      /* ... and we only NULL it after a signal handler can start
+         writing to it.  */
+      sig->next = NULL;
+      barrier ();
+      func (data);
+      sig = next;
+    }
+}
+
 void
 _gst_async_signal (OOP semaphoreOOP)
 {
-  _gst_async_call (do_async_signal, semaphoreOOP);
+  _gst_async_call (_gst_do_async_signal, semaphoreOOP);
 }
 
 void
 _gst_async_signal_and_unregister (OOP semaphoreOOP)
 {
-  _gst_disable_interrupts (false);	/* block out everything! */
-  _gst_async_call (do_async_signal, semaphoreOOP);
-  _gst_async_call (_gst_unregister_oop, semaphoreOOP);
-  _gst_enable_interrupts (false);
+  _gst_async_call (_gst_do_async_signal_and_unregister, semaphoreOOP);
 }
 
 void
-_gst_sync_wait (OOP semaphoreOOP)
+sync_wait_process (OOP semaphoreOOP, OOP processOOP)
 {
   gst_semaphore sem;
+  mst_Boolean isActive;
 
   sem = (gst_semaphore) OOP_TO_OBJ (semaphoreOOP);
+  if (IS_NIL (processOOP))
+    {
+      processOOP = get_active_process ();
+      isActive = true;
+    }
+  else
+    isActive = (processOOP == get_active_process ());
+
   if (TO_INT (sem->signals) <= 0)
     {
-      /* have to suspend, move this to the end of the list */
-      add_last_link (semaphoreOOP, get_active_process ());
-      if (IS_NIL (ACTIVE_PROCESS_YIELD ()))
+      /* Have to suspend.  Prepare return value for #wait and move
+         this process to the end of the list.
+
+         Tweaking the stack top means that this function should only
+	 be called from a primitive.  */
+      SET_STACKTOP (_gst_nil_oop);
+      remove_process_from_list (processOOP);
+      add_last_link (semaphoreOOP, processOOP);
+      if (isActive && IS_NIL (ACTIVE_PROCESS_YIELD ()))
         {
 	  printf ("No runnable process");
 	  activate_process (_gst_prepare_execution_environment ());
@@ -1647,6 +1710,12 @@ _gst_sync_wait (OOP semaphoreOOP)
   /* printf ("wait %O %O\n", semaphoreOOP, sem->firstLink); */
 }
 
+void
+_gst_sync_wait (OOP semaphoreOOP)
+{
+  sync_wait_process (semaphoreOOP, _gst_nil_oop);
+}
+
 OOP
 remove_first_link (OOP semaphoreOOP)
 {
@@ -1813,6 +1882,39 @@ sleep_process (OOP processOOP)
 }
 
 
+mst_Boolean
+would_reschedule_process ()
+{
+  OOP processLists, processListOOP;
+  int priority, activePriority;
+  OOP processOOP;
+  gst_process process;
+  gst_semaphore processList;
+
+  if (!IS_NIL (switch_to_process))
+    return false;
+
+  processOOP = get_scheduled_process ();
+  process = (gst_process) OOP_TO_OBJ (processOOP);
+  activePriority = TO_INT (process->priority);
+  processLists = GET_PROCESS_LISTS ();
+  priority = NUM_OOPS (OOP_TO_OBJ (processLists));
+  do
+    {
+      assert (priority > 0);
+      processListOOP = ARRAY_AT (processLists, priority);
+    }
+  while (is_empty (processListOOP) && --priority >= activePriority);
+
+  processList = (gst_semaphore) OOP_TO_OBJ (processListOOP);
+  return (priority < activePriority
+	  || (priority == activePriority
+	      /* If the same priority, check if the list has the
+		 current process as the sole element.  */
+	      && processList->firstLink == processList->lastLink
+	      && processList->firstLink == processOOP));
+}
+
 OOP
 highest_priority_process (void)
 {
@@ -1899,8 +2001,12 @@ _gst_check_process_state (void)
 	  if (process->myList != processListOOP)
 	    abort ();
 
+#if 0
+	  /* This is false when a process has just finished initializing
+	     itself.  */
 	  if (process->priority != FROM_INT (priority))
 	    abort ();
+#endif
 
           /* Sanity check the last link in the process list.  */
 	  if (IS_NIL (process->nextLink) && processOOP != processList->lastLink)
@@ -1994,9 +2100,6 @@ _gst_init_process_system (void)
     processor->processTimeslice =
       FROM_INT (DEFAULT_PREEMPTION_TIMESLICE);
 
-  if (IS_NIL (processor->gcSemaphore))
-    processor->gcSemaphore = semaphore_new (0);
-
   /* No process is active -- so highest_priority_process() need not
      worry about discarding an active process.  */
   processor->activeProcess = _gst_nil_oop;
@@ -2011,22 +2114,24 @@ create_callin_process (OOP contextOOP)
   OOP processListsOOP;
   gst_processor_scheduler processor;
   gst_process initialProcess;
-  OOP initialProcessOOP, initialProcessListOOP;
+  OOP initialProcessOOP, initialProcessListOOP, nameOOP;
   inc_ptr inc = INC_SAVE_POINTER ();
 
   processor = (gst_processor_scheduler) OOP_TO_OBJ (_gst_processor_oop);
   processListsOOP = processor->processLists;
   initialProcessListOOP = ARRAY_AT (processListsOOP, 4);
 
+  nameOOP = _gst_string_new ("call-in process");
+  INC_ADD_OOP (nameOOP);
+
   initialProcess = (gst_process)
     instantiate (_gst_callin_process_class, &initialProcessOOP);
 
   INC_ADD_OOP (initialProcessOOP);
-
   initialProcess->priority = FROM_INT (USER_SCHEDULING_PRIORITY);
   initialProcess->interruptLock = _gst_nil_oop;
   initialProcess->suspendedContext = contextOOP;
-  initialProcess->name = _gst_string_new ("call-in process");
+  initialProcess->name = nameOOP;
   INC_RESTORE_POINTER (inc);
 
   /* Put initialProcessOOP in the root set */
@@ -2111,12 +2216,6 @@ _gst_init_interpreter (void)
 #endif
 
   _gst_this_context_oop = _gst_nil_oop;
-  async_queue_index = 0;
-  async_queue_index_sig = 0;
-  async_queue_size = 32;
-  queued_async_signals = malloc (sizeof (async_queue_entry) * async_queue_size);
-  gl_lock_init (async_queue_lock);
-
   for (i = 0; i < MAX_LIFO_DEPTH; i++)
     lifo_contexts[i].flags = F_POOLED | F_CONTEXT;
 
@@ -2235,8 +2334,7 @@ set_preemption_timer (void)
 
   time_to_preempt = false;
   if (timeSlice > 0)
-    _gst_signal_after (timeSlice, preempt_smalltalk_process,
-		       TIMER_PROCESS);
+    _gst_sigvtalrm_every (timeSlice, preempt_smalltalk_process);
 #endif
 }
 
@@ -2286,16 +2384,14 @@ _gst_copy_processor_registers (void)
 void
 copy_semaphore_oops (void)
 {
-  int i;
+  async_queue_entry *sig;
 
-  _gst_disable_interrupts (false);	/* block out everything! */
-
-  for (i = 0; i < async_queue_index; i++)
-    if (queued_async_signals[i].data)
-      MAYBE_COPY_OOP (queued_async_signals[i].data);
-  for (i = 0; i < async_queue_index_sig; i++)
-    if (queued_async_signals_sig[i].data)
-      MAYBE_COPY_OOP (queued_async_signals_sig[i].data);
+  for (sig = queued_async_signals; sig != &queued_async_signals_tail;
+       sig = sig->next)
+    MAYBE_COPY_OOP (sig->data);
+  for (sig = queued_async_signals_sig; sig != &queued_async_signals_tail;
+       sig = sig->next)
+    MAYBE_COPY_OOP (sig->data);
 
   /* there does seem to be a window where this is not valid */
   if (single_step_semaphore)
@@ -2303,8 +2399,6 @@ copy_semaphore_oops (void)
 
   /* there does seem to be a window where this is not valid */
   MAYBE_COPY_OOP (switch_to_process);
-
-  _gst_enable_interrupts (false);
 }
 
 
@@ -2324,16 +2418,14 @@ _gst_mark_processor_registers (void)
 void
 mark_semaphore_oops (void)
 {
-  int i;
+  async_queue_entry *sig;
 
-  _gst_disable_interrupts (false);	/* block out everything! */
-
-  for (i = 0; i < async_queue_index; i++)
-    if (queued_async_signals[i].data)
-      MAYBE_MARK_OOP (queued_async_signals[i].data);
-  for (i = 0; i < async_queue_index_sig; i++)
-    if (queued_async_signals_sig[i].data)
-      MAYBE_MARK_OOP (queued_async_signals_sig[i].data);
+  for (sig = queued_async_signals; sig != &queued_async_signals_tail;
+       sig = sig->next)
+    MAYBE_MARK_OOP (sig->data);
+  for (sig = queued_async_signals_sig; sig != &queued_async_signals_tail;
+       sig = sig->next)
+    MAYBE_MARK_OOP (sig->data);
 
   /* there does seem to be a window where this is not valid */
   if (single_step_semaphore)
@@ -2341,8 +2433,6 @@ mark_semaphore_oops (void)
 
   /* there does seem to be a window where this is not valid */
   MAYBE_MARK_OOP (switch_to_process);
-
-  _gst_enable_interrupts (false);
 }
 
 
@@ -2482,12 +2572,14 @@ backtrace_on_signal (int sig)
   raise (sig);
 }
 
+#ifdef SIGUSR1
 static RETSIGTYPE
 user_backtrace_on_signal (int sig)
 {
   _gst_set_signal_handler (sig, user_backtrace_on_signal);
   backtrace_on_signal_1 (false, true);
 }
+#endif
 
 void
 _gst_init_signals (void)
@@ -2501,7 +2593,9 @@ _gst_init_signals (void)
     }
   _gst_set_signal_handler (SIGTERM, backtrace_on_signal);
   _gst_set_signal_handler (SIGINT, interrupt_on_signal);
+#ifdef SIGUSR1
   _gst_set_signal_handler (SIGUSR1, user_backtrace_on_signal);
+#endif
 }
 
 
@@ -2693,13 +2787,31 @@ stop_execution (void)
 }
 
 mst_Boolean
-parse_stream_with_protection (mst_Boolean method)
+parse_method_from_stream_with_protection (OOP currentClass,
+					  OOP currentCategory)
+{
+  interp_jmp_buf jb;
+  OOP methodOOP;
+
+  push_jmp_buf (&jb, false, get_active_process ());
+  if (setjmp (jb.jmpBuf) == 0)
+    methodOOP = _gst_parse_method_from_stream (currentClass,
+                                               currentCategory);
+  else
+    methodOOP = _gst_nil_oop;
+
+  PUSH_OOP (methodOOP);
+  return pop_jmp_buf ();
+}
+
+mst_Boolean
+parse_stream_with_protection (OOP currentNamespace)
 {
   interp_jmp_buf jb;
 
   push_jmp_buf (&jb, false, get_active_process ());
   if (setjmp (jb.jmpBuf) == 0)
-    _gst_parse_stream (method);
+    _gst_parse_stream (currentNamespace);
 
   return pop_jmp_buf ();
 }
