@@ -55,6 +55,33 @@
 #ifndef GST_LEX_H
 #define GST_LEX_H
 
+/* A structure holding a constant for objects having byte-sized
+ *    indexed instance variables (ByteArrays and LargeIntegers).  */
+typedef struct byte_object
+{
+  OOP class;
+  int size;
+  gst_uchar body[1];
+}
+ *byte_object;
+
+typedef union YYSTYPE {
+  long double          fval;
+  intptr_t             ival;
+  char                *sval;
+  byte_object          boval;
+  OOP                  oval;
+  struct tree_node    *node;
+} YYSTYPE;
+
+typedef struct YYLTYPE
+{
+  int first_line;
+  int first_column;
+  int64_t file_offset;
+} YYLTYPE;
+
+
 /* True if errors must be reported to the standard error, false if
    errors should instead stored so that they are passed to Smalltalk
    code.  */
@@ -85,8 +112,13 @@ extern int _gst_first_error_line
 extern struct obstack *_gst_compilation_obstack 
   ATTRIBUTE_HIDDEN;
 
+/* Parse a method from the topmost stream in the stack.  */
+extern OOP _gst_parse_method_from_stream (OOP currentClass,
+					  OOP currentCategory)
+  ATTRIBUTE_HIDDEN;
+
 /* Parse the topmost stream in the stack.  */
-extern void _gst_parse_stream (mst_Boolean method)
+extern void _gst_parse_stream (OOP currentNamespace)
   ATTRIBUTE_HIDDEN;
 
 /* Lexer interface to the lexer.  */
@@ -100,4 +132,10 @@ extern void _gst_yyprint (FILE * file,
 			  PTR lval) 
   ATTRIBUTE_HIDDEN;
 
+/* Negate the semantic value YYLVAL, which must be a numeric token
+   of type TOKEN.  Returns true if YYLVAL is positive, false if it
+   is negative.  */
+extern mst_Boolean _gst_negate_yylval (int token,
+				       YYSTYPE *yylval);
+
 #endif /* GST_LEX_H */
