@@ -166,6 +166,15 @@ static inline mst_Boolean incremental_gc_running (void);
    always passed, but you never know).  */
 static void reset_incremental_gc (OOP firstOOP);
 
+/* Compact the old objects.  Grow oldspace to NEWSIZE bytes.  */
+static void compact (size_t new_heap_limit);
+
+/* Allocate and return space for a fixedspace object of SIZE bytes.
+   The pointer to the object data is returned, the OOP is
+   stored in P_OOP.  */
+static gst_object alloc_fixed_obj (size_t size,
+                                   OOP *p_oop);
+
 /* Gather statistics.  */
 static void update_stats (unsigned long *last, double *between, double *duration);
 
@@ -348,6 +357,9 @@ _gst_init_mem (size_t eden, size_t survivor, size_t old,
       _gst_inc_init_registry ();
     }
 
+  _gst_mem.markQueue = (struct mark_queue *)
+    xcalloc (8 * K, sizeof (struct mark_queue));
+  _gst_mem.lastMarkQueue = &_gst_mem.markQueue[8 * K];
 }
 
 void _gst_update_object_memory_oop (OOP oop)
@@ -472,7 +484,6 @@ alloc_oop_table (size_t size)
   _gst_false_oop = &_gst_mem.ot[FALSE_OOP_INDEX];
 
   _gst_mem.num_free_oops = size;
-  _gst_mem.first_allocated_oop = _gst_mem.ot;
   _gst_mem.last_allocated_oop = _gst_mem.last_swept_oop = _gst_mem.ot - 1;
   _gst_mem.next_oop_to_sweep = _gst_mem.ot - 1;
 }
@@ -763,7 +774,7 @@ _gst_alloc_obj (size_t size,
   newAllocPtr = _gst_mem.eden.allocPtr + BYTES_TO_SIZE (size);
 
   if UNCOMMON (size >= _gst_mem.big_object_threshold)
-    return _gst_alloc_old_obj (size, p_oop);
+    return alloc_fixed_obj (size, p_oop);
 
   if UNCOMMON (newAllocPtr >= _gst_mem.eden.maxPtr)
     {
@@ -780,8 +791,8 @@ _gst_alloc_obj (size_t size,
 }
 
 gst_object
-_gst_alloc_old_obj (size_t size,
-		    OOP *p_oop)
+alloc_fixed_obj (size_t size,
+	         OOP *p_oop)
 {
   gst_object p_instance;
 
@@ -797,7 +808,7 @@ _gst_alloc_old_obj (size_t size,
   if COMMON (p_instance)
     goto ok;
 
-  _gst_compact (0);
+  compact (0);
   p_instance = (gst_object) _gst_mem_alloc (_gst_mem.old, size);
   if UNCOMMON (!p_instance)
     {
@@ -891,23 +902,27 @@ oldspace_before_freeing (heap_data *h, heap_block *blk, size_t sz)
 heap_data *
 oldspace_nomemory (heap_data *h, size_t sz)
 {
+  heap_data **p_heap;
+
+  assert (h == _gst_mem.old || h == _gst_mem.fixed);
+  p_heap = (h == _gst_mem.old ? &_gst_mem.old : &_gst_mem.fixed);
+
   if (!_gst_gc_running)
     _gst_global_gc (sz);
   else
     {
       /* Already garbage collecting, emergency growth just to satisfy
 	 tenuring necessities.  */
-      int grow_amount_to_satisfy_rate = _gst_mem.old->heap_limit
+      int grow_amount_to_satisfy_rate = h->heap_limit
            * (100.0 + _gst_mem.space_grow_rate) / 100;
-      int grow_amount_to_satisfy_threshold = 
-	   (sz + _gst_mem.old->heap_total)
+      int grow_amount_to_satisfy_threshold = (sz + h->heap_total)
 	   * 100.0 /_gst_mem.grow_threshold_percent;
 
-      _gst_mem.old->heap_limit = MAX (grow_amount_to_satisfy_rate,
-				      grow_amount_to_satisfy_threshold);
+      h->heap_limit = MAX (grow_amount_to_satisfy_rate,
+                           grow_amount_to_satisfy_threshold);
     }
 
-  return _gst_mem.old;
+  return *p_heap;
 }
 
 #ifndef NO_SIGSEGV_HANDLING
@@ -972,7 +987,7 @@ update_stats (unsigned long *last, double *between, double *duration)
 void
 _gst_grow_memory_to (size_t spaceSize)
 {
-  _gst_compact (spaceSize);
+  compact (spaceSize);
 }
 
 void
@@ -986,10 +1001,9 @@ grow_memory_no_compact (size_t new_heap_limit)
 }
 
 void
-_gst_compact (size_t new_heap_limit)
+compact (size_t new_heap_limit)
 {
   OOP oop;
-  grey_area_node *node, **next, *last;
   heap_data *new_heap = init_old_space (
     new_heap_limit ? new_heap_limit : _gst_mem.old->heap_limit);
 
@@ -1009,26 +1023,6 @@ _gst_compact (size_t new_heap_limit)
       update_stats (&stats.timeOfLastCompaction, NULL, NULL);
     }
 
-  /* Leave only pages from the loaded image in the grey table.  */
-  for (last = NULL, next = &_gst_mem.grey_pages.head; (node = *next); )
-    if (node->base >= (OOP *)_gst_mem.loaded_base
-        && node->base < _gst_mem.loaded_end)
-      {
-#ifdef MMAN_DEBUG_OUTPUT
-        printf ("  Remembered table entry left for loaded image: %p..%p\n",
-                node->base, node->base+node->n);
-#endif
-        last = node;
-        next = &(node->next);
-      }
-    else
-      {
-        _gst_mem.rememberedTableEntries--;
-        *next = node->next;
-        xfree (node);
-      }
-
-  _gst_mem.grey_pages.tail = last;
   _gst_fixup_object_pointers ();
 
   /* Now do the copying loop which will compact oldspace.  */
@@ -1062,7 +1056,7 @@ void
 _gst_global_compact ()
 {
   _gst_global_gc (0);
-  _gst_compact (0);
+  compact (0);
 }
 
 void
@@ -1137,7 +1131,7 @@ _gst_global_gc (int next_allocation)
 	  if (target_limit < old_limit)
             {
               s = "done, heap compacted";
-              _gst_compact (0);
+              compact (0);
               grow_memory_no_compact (target_limit);
             }
         }
@@ -1400,21 +1394,21 @@ reset_incremental_gc (OOP firstOOP)
       ;
 #endif
 
-  _gst_mem.first_allocated_oop = oop;
+  /* Initialize these here so that IS_OOP_VALID works correctly.  */
+  _gst_mem.next_oop_to_sweep = _gst_mem.last_allocated_oop;
+  _gst_mem.last_swept_oop = oop - 1;
 
 #ifdef NO_INCREMENTAL_GC
-  _gst_mem.next_oop_to_sweep = _gst_mem.last_allocated_oop;
   _gst_finish_incremental_gc ();
 #else
   /* Skip high OOPs that are unallocated.  */
-  for (oop = _gst_mem.last_allocated_oop; !IS_OOP_VALID_GC (oop); oop--)
+  for (oop = _gst_mem.last_allocated_oop; !IS_OOP_VALID (oop); oop--)
     _gst_sweep_oop (oop);
 
   _gst_mem.last_allocated_oop = oop;
   _gst_mem.next_oop_to_sweep = oop;
 #endif
 
-  _gst_mem.last_swept_oop = _gst_mem.first_allocated_oop - 1;
   _gst_mem.num_free_oops = _gst_mem.ot_size -
     (_gst_mem.last_allocated_oop - _gst_mem.ot);
 
@@ -1425,9 +1419,9 @@ reset_incremental_gc (OOP firstOOP)
       * (100 + _gst_mem.space_grow_rate) / 100);
 
 #if defined (GC_DEBUG_OUTPUT)
-  printf ("Found unallocated at OOP %p, last allocated OOP %p\n"
-          "Next OOP swept top to bottom %p, lowest swept bottom to top %p\n",
-	  _gst_mem.first_allocated_oop, _gst_mem.last_allocated_oop,
+  printf ("Last allocated OOP %p\n"
+          "Next OOP swept top to bottom %p, highest swept bottom to top %p\n",
+	  _gst_mem.last_allocated_oop,
 	  _gst_mem.next_oop_to_sweep, _gst_mem.last_swept_oop);
 #endif
 }
@@ -1495,7 +1489,12 @@ mourn_objects (void)
       array = new_instance_with (_gst_array_class, size, &processor->gcArray);
       _gst_copy_buffer (array->data);
       if (!IS_NIL (processor->gcSemaphore))
-        _gst_async_signal (processor->gcSemaphore);
+        {
+          static async_queue_entry e;
+          e.func = _gst_do_async_signal;
+          e.data = processor->gcSemaphore;
+          _gst_async_call_internal (&e);
+        }
       else
 	{
 	  _gst_errorf ("Running finalizers before initialization.");
@@ -1709,8 +1708,8 @@ check_weak_refs ()
       if (!IS_OOP_VALID_GC (oop))
 	continue;
 
-      for (field = (OOP *) oop->object, n = TO_INT (oop->object->objSize);
-	   n--; field++)
+      for (field = (OOP *) oop->object + OBJ_HEADER_SIZE_WORDS,
+	   n = NUM_OOPS (oop->object); n--; field++)
         {
 	  OOP oop = *field;
           if (IS_INT (oop))
@@ -1928,15 +1927,10 @@ int
 scanned_fields_in (gst_object object,
 		  int flags)
 {
-  OOP objClass = object->objClass;
-
   if COMMON (!(flags & (F_WEAK | F_CONTEXT)))
     {
       int size = NUM_OOPS (object);
-      if COMMON (size)
-	return object->data + size - &object->objClass;
-      else
-	return UNCOMMON (!IS_OOP_COPIED (objClass));
+      return object->data + size - &object->objClass;
     }
 
   if COMMON (flags & F_CONTEXT)
@@ -1947,14 +1941,9 @@ scanned_fields_in (gst_object object,
       methodSP = TO_INT (ctx->spOffset);
       return ctx->contextStack + methodSP + 1 - &ctx->objClass;
     }
-  else
-    {
-      /* In general, there will be many instances of a class,
-	 but only the first time will it need to be copied;
-         moreover, classes are often old.  So I'm
-	 marking this as uncommon.  */
-      return UNCOMMON (!IS_OOP_COPIED (objClass));
-    }
+
+  /* Weak object, only mark the class.  */
+  return 1;
 }
 
 void
@@ -2181,50 +2170,70 @@ void
 _gst_mark_an_oop_internal (OOP oop)
 {
   OOP *curOOP, *atEndOOP;
+  struct mark_queue *markQueue = _gst_mem.markQueue;
+  struct mark_queue *lastMarkQueue = _gst_mem.lastMarkQueue;
+  struct mark_queue *currentMarkQueue = markQueue;
   goto markOne;
 
  markRange:
-  {			/* in the loop! */
-#if defined (GC_DEBUGGING)
-    gst_object obj = (gst_object) (curOOP - 1);	/* for debugging */
-#endif
-    for (;;)
+  {
+    OOP firstOOP = NULL;
+
+    /* The first unmarked OOP is used for tail recursion.  */
+    while (curOOP < atEndOOP)
       {
-        /* in a loop, do next iteration */
-        oop = *curOOP;
-        PREFETCH_LOOP (curOOP, PREF_READ | PREF_NTA);
-        curOOP++;
-        if (IS_OOP (oop))
+        oop = *curOOP++;
+        if (IS_OOP (oop) && !IS_OOP_MARKED (oop))
           {
-#if defined (GC_DEBUGGING)
-            if UNCOMMON (!IS_OOP_ADDR (oop))
+            oop->flags |= F_REACHABLE;
+            firstOOP = oop;
+            break;
+          }
+      }
+
+    /* The second unmarked OOP is the first that is placed on the mark
+       queue.  TODO: split REACHABLE and VISITED flags.  An object is
+       marked REACHABLE here, and REACHABLE|VISITED in the markOne label.
+       At the end of GC, all REACHABLE objects are also VISITED.
+       The above loop should seek an object that is not VISITED so
+       that it can be marked.  For the loop below, however, REACHABLE
+       objects are known to be somewhere else on the mark stack, and can
+       be skipped.
+
+       Skipping objects after the first unmarked OOP is still useful,
+       because it keeps the stack size a bit lower in the relatively common
+       case of many integer or nil instance variables.  */
+    while (curOOP < atEndOOP)
+      {
+        oop = *curOOP;
+
+        if (IS_OOP (oop) && !IS_OOP_MARKED (oop))
+          {
+            if (currentMarkQueue == lastMarkQueue)
               {
-                printf
-                  ("Error! Invalid OOP %p was found inside %p!\n",
-                   oop, obj);
-                abort ();
+                const size_t size = lastMarkQueue - markQueue;
+
+                _gst_mem.markQueue = (struct mark_queue *)
+                  xrealloc (_gst_mem.markQueue, 2 * size * sizeof (struct mark_queue));
+                _gst_mem.lastMarkQueue = &_gst_mem.markQueue[2 * size];
+
+		markQueue = _gst_mem.markQueue;
+		lastMarkQueue = _gst_mem.lastMarkQueue;
+                currentMarkQueue = &_gst_mem.markQueue[size];
               }
-            else
-#endif
-              if (!IS_OOP_MARKED (oop))
-                {
-                  PREFETCH_START (oop->object, PREF_READ | PREF_NTA);
-
-                  /* On the last object in the set, reuse the
-                     current invocation. oop is valid, so we go to
-                     the single-object case */
-                  if UNCOMMON (curOOP == atEndOOP)
-                    goto markOne;
-
-                  _gst_mark_an_oop_internal (oop);
-                  continue;
-                }
+            currentMarkQueue->firstOOP = curOOP;
+            currentMarkQueue->endOOP = atEndOOP;
+	    currentMarkQueue++;
+            break;
           }
 
-        /* Place the check here so that the continue above skips it.  */
-        if UNCOMMON (curOOP == atEndOOP)
-          return;
+        curOOP++;
       }
+
+    if (!firstOOP)
+      goto pop;
+
+    TAIL_MARK_OOP (firstOOP);
   }
 
  markOne:
@@ -2286,6 +2295,16 @@ _gst_mark_an_oop_internal (OOP oop)
           TAIL_MARK_OOP (objClass);
       }
   }
+
+ pop:
+  {
+    if (currentMarkQueue > markQueue)
+      {
+        currentMarkQueue--;
+	TAIL_MARK_OOPRANGE (currentMarkQueue->firstOOP,
+			    currentMarkQueue->endOOP);
+      }
+  }
 }
 
 void
