@@ -315,7 +315,7 @@ void
 save_to_fd (int imageFd)
 {
   save_file_header header;
-  memzero (&header, sizeof (header));
+  memset (&header, 0, sizeof (header));
   myOOPTable = make_oop_table_to_be_saved (&header);
 
   buffer_write_init (imageFd, WRITE_BUFFER_SIZE);
@@ -380,15 +380,6 @@ make_oop_table_to_be_saved (struct save_file_header *header)
 	  int numPointers = NUM_OOPS (oop->object);
 
           myOOPTable[i].flags = (oop->flags & ~F_RUNTIME) | F_OLD;
-
-	  /* Cache the number of indexed instance variables.  We prefer
-	     to do more work upon saving (done once) than upon loading
-	     (done many times).  */
-	  if (numPointers < (F_COUNT >> F_COUNT_SHIFT))
-	    myOOPTable[i].flags |= numPointers << F_COUNT_SHIFT;
-	  else
-	    myOOPTable[i].flags |= F_COUNT;
-
 	  myOOPTable[i].object = (gst_object) TO_INT (oop->object->objSize);
 	}
       else
@@ -765,6 +756,13 @@ fixup_object (OOP oop, gst_object dest, gst_object src, int numBytes)
 	}
     }
 
+  /* File descriptors are invalidated on resume.  */
+  else if (is_a_kind_of (class_oop, _gst_file_descriptor_class))
+    {
+      gst_file_stream file = (gst_file_stream) dest;
+      file->fd = _gst_nil_oop;
+    }
+
   /* The other case is to reset CFunctionDescriptor objects, so that we'll
      relink the external functions when we reload the image.  */
   else if (is_a_kind_of (class_oop, _gst_c_callable_class))
@@ -797,11 +795,7 @@ restore_oop_pointer_slots (OOP oop)
   object = OOP_TO_OBJ (oop);
   object->objClass = OOP_ABSOLUTE (object->objClass);
 
-  if UNCOMMON ((oop->flags & F_COUNT) == F_COUNT)
-    numPointers = NUM_OOPS (object);
-  else
-    numPointers = oop->flags >> F_COUNT_SHIFT;
-
+  numPointers = NUM_OOPS (object);
   for (i = object->data; numPointers--; i++)
     if (IS_OOP (*i))
       *i = OOP_ABSOLUTE (*i);
