@@ -7,8 +7,8 @@
 
 /***********************************************************************
  *
- * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2003,2005,2006,2007,2008
- * Free Software Foundation, Inc.
+ * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2003,
+ * 2005,2006,2007,2008,2009 Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
  * This file is part of GNU Smalltalk.
@@ -77,7 +77,6 @@ struct symbol_list
    which are declared in a given scope.  Nested scopes result in
    nested instances of the scope struct, with the current scope always
    being the innermost at any point during the compilation.  */
-typedef struct scope *scope;
 struct scope
 {
   scope prevScope;
@@ -89,7 +88,6 @@ struct scope
 /* Represents all the pools (namespaces) which are declared in the
    current scope.  This information is relatively complex to compute,
    so it's kept cached.  */
-typedef struct pool_list *pool_list;
 struct pool_list
 {
   OOP poolOOP;
@@ -179,9 +177,6 @@ OOP temporaries_dictionary = NULL;
 /* The list of selectors for the send immediate bytecode.  */
 struct builtin_selector _gst_builtin_selectors[256] = {};
 
-/* True if undeclared globals can be considered forward references.  */
-enum undeclared_strategy _gst_use_undeclared = UNDECLARED_TEMPORARIES;
-
 /* Answer whether OOP is a Smalltalk String LEN characters long and
    these characters match the first LEN characters of STR (which must
    not have embedded NULs).  */
@@ -250,12 +245,6 @@ intern_string_fast (const char *str, OOP *pTestOOP);
    found.  */
 static int instance_variable_index (OOP symbol);
 
-/* This checks if the INDEX-th instance variable among those that the
-   current class declares is read-only.  Read-only index variables are
-   those that are declared by a trusted super-class of an untrusted
-   subclass.  */
-static mst_Boolean is_instance_variable_read_only (int index);
-
 /* This looks for SYMBOL among the arguments and temporary variables
    that the current scope sees, and returns the entry in the symbol
    list for the variable if it is found.  */
@@ -267,9 +256,6 @@ static symbol_list find_local_var (scope scope,
    the symbol list for the variable if it is found.  */
 static OOP find_class_variable (OOP varName);
 
-static scope cur_scope = NULL;
-static pool_list linearized_pools = NULL;
-
 /* This is an array of symbols which the virtual machine knows about,
    and is used to restore the global variables upon image load.  */
 static const symbol_info sym_info[] = {
@@ -353,13 +339,13 @@ _gst_get_scope_kind (scope_type scope)
 int
 _gst_get_arg_count (void)
 {
-  return (cur_scope->numArguments);
+  return (_gst_compiler_state->cur_scope->numArguments);
 }
 
 int
 _gst_get_temp_count (void)
 {
-  return (cur_scope->numTemporaries);
+  return (_gst_compiler_state->cur_scope->numTemporaries);
 }
 
 void
@@ -367,11 +353,11 @@ _gst_push_new_scope (void)
 {
   scope newScope;
   newScope = (scope) xmalloc (sizeof (*newScope));
-  newScope->prevScope = cur_scope;
+  newScope->prevScope = _gst_compiler_state->cur_scope;
   newScope->symbols = NULL;
   newScope->numArguments = 0;
   newScope->numTemporaries = 0;
-  cur_scope = newScope;
+  _gst_compiler_state->cur_scope = newScope;
 }
 
 void
@@ -379,8 +365,8 @@ _gst_pop_old_scope (void)
 {
   scope oldScope;
 
-  oldScope = cur_scope;
-  cur_scope = cur_scope->prevScope;
+  oldScope = _gst_compiler_state->cur_scope;
+  _gst_compiler_state->cur_scope = oldScope->prevScope;
 
   free_scope_symbols (oldScope);
   xfree (oldScope);
@@ -388,17 +374,21 @@ _gst_pop_old_scope (void)
 
 void
 _gst_pop_all_scopes (void)
+{
+  while (_gst_compiler_state->cur_scope)
+    _gst_pop_old_scope ();
+}
+
+void
+_gst_free_linearized_pools ()
 {
   pool_list next;
 
-  while (cur_scope)
-    _gst_pop_old_scope ();
-
-  while (linearized_pools)
+  while (_gst_current_parser->linearized_pools)
     {
-      next = linearized_pools->next;
-      xfree (linearized_pools);
-      linearized_pools = next;
+      next = _gst_current_parser->linearized_pools->next;
+      xfree (_gst_current_parser->linearized_pools);
+      _gst_current_parser->linearized_pools = next;
     }
 }
 
@@ -406,6 +396,8 @@ _gst_pop_all_scopes (void)
 int
 _gst_declare_arguments (tree_node args)
 {
+  scope cur_scope = _gst_compiler_state->cur_scope;
+
   if (args->nodeType == TREE_UNARY_EXPR)
     return (0);
 
@@ -440,6 +432,8 @@ _gst_declare_temporaries (tree_node temps)
 int
 _gst_declare_block_arguments (tree_node args)
 {
+  scope cur_scope = _gst_compiler_state->cur_scope;
+
   for (; args != NULL; args = args->v_list.next)
     if (_gst_declare_name (args->v_list.name, false, false) == -1)
       return -1;
@@ -454,6 +448,7 @@ void
 _gst_undeclare_name (void)
 {
   symbol_list oldList;
+  scope cur_scope = _gst_compiler_state->cur_scope;
 
   oldList = cur_scope->symbols;
   cur_scope->symbols = cur_scope->symbols->prevSymbol;
@@ -468,6 +463,7 @@ _gst_declare_name (const char *name,
 {
   symbol_list newList;
   OOP symbol = _gst_intern_string (name);
+  scope cur_scope = _gst_compiler_state->cur_scope;
 
   if (!allowDup && find_local_var (cur_scope, symbol) != NULL)
     return -1;
@@ -805,28 +801,22 @@ combine_local_pools (OOP sharedPoolsOOP, struct pointer_set_t *white, pool_list
    the superclass.  */
 
 static pool_list *
-add_local_pool_resolution (OOP class_oop, pool_list *p_end)
+add_shared_pool_resolution (OOP class_oop, OOP environmentOOP, pool_list *p_end)
 {
-  OOP environmentOOP;
   gst_class class;
   struct pointer_set_t *pset;
 
-  /* First search in the class pool.  */
-  p_end = add_pool (_gst_class_variable_dictionary (class_oop), p_end);
-
   /* Then in all the imports not reachable from the environment.  */
-  class = (gst_class) OOP_TO_OBJ (class_oop);
-  environmentOOP = class->environment;
   pset = make_with_all_superspaces_set (environmentOOP);
+  class = (gst_class) OOP_TO_OBJ (class_oop);
   p_end = combine_local_pools (class->sharedPools, pset, p_end);
   pointer_set_destroy (pset);
 
   /* Then search in the `environments', except those that are already
      reachable from the superclass. */
   class_oop = SUPERCLASS (class_oop);
-  class = (gst_class) OOP_TO_OBJ (class_oop);
   if (!IS_NIL (class_oop))
-    pset = make_with_all_superspaces_set (class->environment);
+    pset = make_with_all_superspaces_set (CLASS_ENVIRONMENT (class_oop));
   else
     pset = NULL;
   
@@ -836,24 +826,42 @@ add_local_pool_resolution (OOP class_oop, pool_list *p_end)
   return p_end;
 }
 
+void
+_gst_compute_linearized_pools (gst_parser *parser, mst_Boolean forDoit)
+{
+  pool_list *p_end = &_gst_current_parser->linearized_pools;
+  OOP myClass, classOOP;
+  OOP environmentOOP = parser->current_namespace;
+
+  if (IS_NIL (parser->currentClass))
+    myClass = _gst_undefined_object_class;
+  else
+    myClass = _gst_get_class_object (parser->currentClass);
+
+  assert (_gst_current_parser->linearized_pools == NULL);
+
+  /* Add pools separately for each class.  */
+  for (classOOP = myClass; !IS_NIL (classOOP); classOOP = SUPERCLASS (classOOP))
+    {
+      /* First search in the class pool.  */
+      p_end = add_pool (_gst_class_variable_dictionary (classOOP), p_end);
+
+      /* Override CLASSOOP's environment with ENVIRONMENTOOP if it is not
+         NULL.  */
+      p_end = add_shared_pool_resolution
+	      (classOOP, 
+	       forDoit ? environmentOOP : CLASS_ENVIRONMENT (classOOP),
+	       p_end);
+    }
+}
+
 OOP
 find_class_variable (OOP varName)
 {
   pool_list pool;
   OOP assocOOP;
 
-  if (!linearized_pools)
-    {
-      pool_list *p_end = &linearized_pools;
-      OOP myClass;
-
-      /* Add pools separately for each class.  */
-      for (myClass = _gst_get_class_object (_gst_this_class); !IS_NIL (myClass);
-           myClass = SUPERCLASS (myClass))
-        p_end = add_local_pool_resolution (myClass, p_end);
-    }
-
-  for (pool = linearized_pools; pool; pool = pool->next)
+  for (pool = _gst_current_parser->linearized_pools; pool; pool = pool->next)
     {
       assocOOP =
 	dictionary_association_at (pool->poolOOP, varName);
@@ -866,15 +874,6 @@ find_class_variable (OOP varName)
 }
 
 
-int
-_gst_set_undeclared (enum undeclared_strategy new)
-{
-  enum undeclared_strategy old = _gst_use_undeclared;
-  if (new != UNDECLARED_CURRENT)
-    _gst_use_undeclared = new;
-  return old;
-}
-
 OOP
 _gst_push_temporaries_dictionary (void)
 {
@@ -911,9 +910,14 @@ _gst_find_variable_binding (tree_node list)
   if (!IS_NIL (assocOOP))
     return _gst_make_oop_constant (&list->location, assocOOP);
 
-  else if (_gst_use_undeclared == UNDECLARED_GLOBALS
-	   && !elt->v_list.next 
-	   && isupper (*STRING_OOP_CHARS (symbol)))
+  /* For temporaries, make a deferred binding so that we can try using
+     a global variable.  Unlike namespaces, the temporaries dictionary
+     does not know anything about Undeclared.  */
+  if (_gst_compiler_state->undeclared_temporaries)
+    return _gst_make_deferred_binding_constant (&list->location, list);
+
+  if (!elt->v_list.next 
+      && isupper (*STRING_OOP_CHARS (symbol)))
     {
       OOP dictOOP = dictionary_at (_gst_smalltalk_dictionary,
 				   _gst_undeclared_symbol);
@@ -923,20 +927,13 @@ _gst_find_variable_binding (tree_node list)
       return _gst_make_oop_constant (&list->location, assocOOP);
     }
 
-  /* For temporaries, make a deferred binding so that we can try using
-     a global variable.  Unlike namespaces, the temporaries dictionary
-     does not know anything about Undeclared.  */
-  else if (_gst_use_undeclared == UNDECLARED_TEMPORARIES)
-    return _gst_make_deferred_binding_constant (&list->location, list);
-
-  else
-    return NULL;
+  return NULL;
 }
 
 OOP
 _gst_get_undeclared_dictionary ()
 {
-  assert (_gst_use_undeclared == UNDECLARED_TEMPORARIES);
+  assert (_gst_compiler_state->undeclared_temporaries);
   return temporaries_dictionary;
 }
 
@@ -982,7 +979,7 @@ _gst_find_variable (symbol_entry * se,
       return (true);
     }
 
-  for (scope = cur_scope, scopeDistance = 0; scope != NULL;
+  for (scope = _gst_compiler_state->cur_scope, scopeDistance = 0; scope != NULL;
        scope = scope->prevScope, scopeDistance++)
     {
       s = find_local_var (scope, symbol);
@@ -999,7 +996,7 @@ _gst_find_variable (symbol_entry * se,
   if (index >= 0)
     {
       fill_symbol_entry (se, SCOPE_RECEIVER, 
-			 is_instance_variable_read_only (index),
+			 false,
 			 symbol, index, 0);
       return (true);
     }
@@ -1012,35 +1009,19 @@ _gst_find_variable (symbol_entry * se,
   index = _gst_add_forced_object (varAssoc);
 
   fill_symbol_entry (se, SCOPE_GLOBAL, 
-		     _gst_untrusted_methods && !IS_OOP_UNTRUSTED (varAssoc),
+		     false,
 		     varAssoc, index, 0);
   return (true);
 }
 
-static mst_Boolean
-is_instance_variable_read_only (int index)
-{
-  int numVars;
-  OOP class_oop;
-
-  if (!_gst_untrusted_methods)
-    return (false);
-
-  for (class_oop = _gst_this_class; IS_OOP_UNTRUSTED (class_oop);
-       class_oop = SUPERCLASS (class_oop))
-    ;
-
-  numVars = CLASS_FIXED_FIELDS (class_oop);
-  return index + 1 <= numVars;
-}
-
 static int
 instance_variable_index (OOP symbol)
 {
   OOP arrayOOP;
   int index, numVars;
 
-  arrayOOP = _gst_instance_variable_array (_gst_this_class);
+  arrayOOP = _gst_instance_variable_array
+    (_gst_curr_method->v_method.currentClass);
   numVars = NUM_OOPS (OOP_TO_OBJ (arrayOOP));
 
   for (index = numVars; index >= 1; index--)
@@ -1114,7 +1095,7 @@ _gst_find_pragma_handler (OOP classOOP,
 {
   OOP class_oop, myClass;
 
-  myClass = _gst_get_class_object (_gst_this_class);
+  myClass = _gst_get_class_object (_gst_curr_method->v_method.currentClass);
 
   /* Now search in the class pools */
   for (class_oop = myClass; !IS_NIL (class_oop);
@@ -1206,7 +1187,7 @@ _gst_make_class_variable_dictionary (const char *variableNames,
   inc_ptr incPtr;
 
   if (variableNames == NULL)
-    variableNames = (gst_uchar *) "";
+    variableNames = "";
 
   incPtr = INC_SAVE_POINTER ();
 
@@ -1497,27 +1478,6 @@ _gst_hash_string (const char *str,
   return hashVal & MAX_ST_INT;
 }
 
-char *
-_gst_symbol_as_string (OOP symbolOOP)
-{
-  static char stringBuf[256];	/* probably large enough for most
-				   symbols */
-  unsigned int len;
-  gst_symbol symbol;
-
-  symbol = (gst_symbol) OOP_TO_OBJ (symbolOOP);
-
-  len = _gst_string_oop_len (symbolOOP);
-  if (len >= sizeof (stringBuf))
-    _gst_errorf ("symbol name too long: %d, max is %d", len,
-		 sizeof (stringBuf));
-
-  strncpy (stringBuf, symbol->symString, len);
-  stringBuf[len] = '\0';
-  return (stringBuf);
-}
-
-
 
 void
 _gst_check_symbol_chain (void)
@@ -1577,15 +1537,16 @@ _gst_selector_num_args (OOP symbolOOP)
 
   len = _gst_string_oop_len (symbolOOP);
   bytes = (char *) (OOP_TO_OBJ (symbolOOP)->data);
-  if (bytes[0] < 'A' || bytes[0] > 'z')
-    return (1);
-
-  if (bytes[0] > 'Z' && bytes[0] < 'a')
-    return (1);
-
-  for (numArgs = 0; len;)
-    if (bytes[--len] == ':')
-      numArgs++;
+  if ((bytes[0] >= 'A' && bytes[0] <= 'Z')
+      || (bytes[0] >= 'a' && bytes[0] <= 'z')
+      || bytes[0] == '_')
+    {
+      for (numArgs = 0; len;)
+        if (bytes[--len] == ':')
+          numArgs++;
+    }
+  else
+    numArgs = 1;
 
   return (numArgs);
 }
